import os
import sys
import itertools
import glob
from typing import Union

# import the different readers
from resistics.ioHandlers.dataReader import DataReader
from resistics.ioHandlers.dataReaderATS import DataReaderATS
from resistics.ioHandlers.dataReaderSpam import DataReaderSPAM
from resistics.ioHandlers.dataReaderInternal import DataReaderInternal
from resistics.ioHandlers.dataReaderAscii import DataReaderAscii
from resistics.ioHandlers.dataReaderPhoenix import DataReaderPhoenix


def getDataReader(datapath: str) -> Union[DataReader, bool]:
    """Get the data reader for a time data directory (and format)
    
    Parameters
    ----------
    datapath : str
        Path to data directory

    Returns
    -------
    dataReader : DataReader (DataReaderATS, DataReaderSPAM, DataReaderInternal, DataReaderPhonix)
        Data reader object
    """
    
    # check the file endings in the datapath
    # internal format
    headerF = glob.glob(os.path.join(datapath, "*.hdr"))
    headerF = headerF + glob.glob(os.path.join(datapath, "*.HDR"))
    if len(headerF) > 0:  # then internal format or ascii format
        dataF = glob.glob(os.path.join(datapath, "*.ascii"))
        if len(dataF) > 0:
            return DataReaderAscii(datapath)
        # otherwise assume internal format
        return DataReaderInternal(datapath)
    # ATS data files
    headerF = glob.glob(os.path.join(datapath, "*.xml"))
    headerF = headerF + glob.glob(os.path.join(datapath, "*.XML"))
    if len(headerF) > 0:  # then ats format
        return DataReaderATS(datapath)
    # SPAM data files
    headerF = glob.glob(os.path.join(datapath, "*.xtrx"))
    headerF = headerF + glob.glob(os.path.join(datapath, "*.XTRX"))
    if len(headerF) > 0:  # then xtrx format
        return DataReaderSPAM(datapath)
    headerF = glob.glob(os.path.join(datapath, "*.xtr"))
    headerF = headerF + glob.glob(os.path.join(datapath, "*.XTR"))
    if len(headerF) > 0:  # then xtr format
        return DataReaderSPAM(datapath)
    # Phoenix data files
    headerF = glob.glob(os.path.join(datapath, "*.tbl"))
    headerF = headerF + glob.glob(os.path.join(datapath, "*.TBL"))
    if len(headerF) > 0:  # then xtr format
        return DataReaderPhoenix(datapath)
    # if nothing found, return false
    return False



from copy import deepcopy
from datetime import datetime, timedelta
import numpy as np
import matplotlib.pyplot as plt
from typing import List, Dict, Union

# import from package
from resistics.dataObjects.dataObject import DataObject
from resistics.utilities.utilsChecks import isElectric
from resistics.utilities.utilsFreq import getFrequencyArray
from resistics.utilities.utilsPlotter import getViewFonts


class SpectrumData(DataObject):
    """Class for holding spectra data

    Attributes
    ----------
    windowSize : int
        The size of the time window in samples
    dataSize : int
        The number of samples in the frequency spectra
    sampleFreq : float
        The sampling frequency
    startTime : datetime.datetime
        The time of the first sample of the time data which was fourier transformed
    stopTime : datetime.time
        The time of the last sample of the time data which was fourier transformed
    chans : List[str]
        The channels in the data
    numChans : int
        The number of channels
    data : Dict
        The spectra data with channels as keys and arrays as values
    comments : List[str]
        Information about the spectra data as a list of strings

    Methods
    -------
    __init__(kwargs)
        Initialise spectra data
    setData(windowSize, dataSize, sampleFreq, startTime, stopTime, data)
        Set data with parameters
    getComments()
        Get a deepcopy of the comments        
    addComment(comment)
        Add a comment to the dataset
    copy()
        Get a copy of the spectrum data
    view(kwargs)
        View the spectra data 
    printList()
        Class status returned as list of strings          
    """

    def __init__(
        self,
        windowSize: int,
        dataSize: int,
        sampleFreq: float,
        startTime: Union[datetime, str],
        stopTime: Union[datetime, str],
        data,
        comments: List[str] = [],
    ) -> None:
        """Initialise and set object parameters

        Parameters
        ----------
        windowSize : int
            The window size in samples of the time data
        dataSize : int
            The spectra size in samples
        sampleFreq : float
            The sampling frequency in Hz
        startTime : datetime, str
            The startTime of the window
        stopTime : datetime, str
            The stopTime of the window 
        data : Dict
            The data dictionary with keys as channels and values as spectra data
        comments : List[str]
            Dataset comments
        """

        self.setData(windowSize, dataSize, sampleFreq, startTime, stopTime, data)
        self.comments = comments

    def setData(
        self,
        windowSize: int,
        dataSize: int,
        sampleFreq: float,
        startTime: Union[datetime, str],
        stopTime: Union[datetime, str],
        data,
    ) -> None:
        """Set the object parameters

        Parameters
        ----------
        windowSize : int
            The window size in samples of the time data
        dataSize : int
            The spectra size in samples
        sampleFreq : float
            The sampling frequency in Hz
        startTime : datetime, str
            The startTime of the window
        stopTime : datetime, str
            The stopTime of the window 
        data : Dict
            The data dictionary with keys as channels and values as spectra data
        comments : List[str]
            Dataset comments
        """

        self.windowSize = windowSize
        self.sampleFreq = sampleFreq
        # start time
        self.startTime = (
            datetime.strptime(startTime, "%Y-%m-%d %H:%M:%S.%f")
            if isinstance(startTime, str)
            else startTime
        )
        # stop time
        self.stopTime = (
            datetime.strptime(stopTime, "%Y-%m-%d %H:%M:%S.%f")
            if isinstance(stopTime, str)
            else stopTime
        )
        # other properties
        self.chans = sorted(data.keys())
        self.numChans = len(self.chans)
        self.data = data
        self.dataSize = data[self.chans[0]].size

    @property
    def nyquist(self) -> float:
        """Get the nyquist frequency of the spectra data

        Returns
        -------
        nyquist : float
            The nyquist frequency in Hz
        """

        return self.sampleFreq / 2.0

    @property
    def freqArray(self) -> np.ndarray:
        """Get the frequency array of the spectra data

        Returns
        -------
        freqArray : np.ndarray
            Array of frequencies
        """

        return getFrequencyArray(self.sampleFreq, self.dataSize)

    def getComments(self) -> List[str]:
        """Get a deepcopy of the comments
        
        Returns
        -------
        List[str]
            Dataset comments as a list of strings
        """

        return deepcopy(self.comments)

    def addComment(self, comment: str) -> None:
        """Add a new comment

        Parameters
        ----------
        comment : float
            A new comment
        """

        self.comments.append(comment)

    def copy(self):
        """Get a copy of the time data object

        Returns
        -------
        TimeData
            A copy of the time data object
        """
        
        return SpectrumData(
            self.windowSize,
            self.dataSize, 
            self.sampleFreq,
            self.startTime,
            self.stopTime,
            deepcopy(self.data),
            self.getComments(),
        )

    def view(self, **kwargs) -> plt.figure:
        """Plot spectra data

        Parameters
        ----------
        fig : matplotlib.pyplot.figure, optional
            A figure object
        plotfonts : Dict, optional
            A dictionary of plot fonts
        chans : List[str], optional
            A list of channels to plot
        label : str, optional
            Label for the plots
        xlim : List, optional
            Limits for the x axis
        color : str, rgba Tuple
            The color for the line plot
        legend : bool
            Boolean flag for adding a legend

        Returns
        -------
        plt.figure
            Matplotlib figure object
        """

        f = self.freqArray
        fig = (
            plt.figure(kwargs["fig"].number)
            if "fig" in kwargs
            else plt.figure(figsize=(20, 2 * self.numChans))
        )
        plotFonts = kwargs["plotfonts"] if "plotfonts" in kwargs else getViewFonts()
        color = kwargs["color"] if "color" in kwargs else None
        # suptitle
        st = fig.suptitle(
            "Spectra data from {} to {}".format(self.startTime, self.stopTime),
            fontsize=plotFonts["suptitle"],
        )
        st.set_y(0.98)
        # now plot the data
        dataChans = kwargs["chans"] if "chans" in kwargs else self.chans
        numPlotChans = len(dataChans)
        for idx, chan in enumerate(dataChans):
            ax = plt.subplot(numPlotChans, 1, idx + 1)
            plt.title("Channel {}".format(chan), fontsize=plotFonts["title"])
            # plot the data
            if "label" in kwargs:
                plt.plot(f, np.absolute(self.data[chan]), color=color, label=kwargs["label"])
            else:
                plt.plot(f, np.absolute(self.data[chan]), color=color)
            # add frequency label
            if idx == numPlotChans - 1:
                plt.xlabel("Frequency [Hz]", fontsize=plotFonts["axisLabel"])
            # x axis options
            xlim = kwargs["xlim"] if "xlim" in kwargs else [f[0], f[-1]]
            plt.xlim(xlim)
            # y axis options
            if isElectric(chan):
                plt.ylabel("[mV/km]", fontsize=plotFonts["axisLabel"])
            else:
                plt.ylabel("[nT]", fontsize=plotFonts["axisLabel"])
            plt.grid()
            # set tick sizes
            for label in ax.get_xticklabels() + ax.get_yticklabels():
                label.set_fontsize(plotFonts["axisTicks"])
            # legend
            if "legend" in kwargs and kwargs["legend"]:
                plt.legend(loc=4)

        # show if the figure is not in keywords
        if "fig" not in kwargs:
            plt.tight_layout(rect=[0, 0.02, 1, 0.96])
            plt.show()
        
        return fig

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Sample rate [s] = {}".format(1.0 / self.sampleFreq))
        textLst.append("Sampling frequency [Hz] = {}".format(self.sampleFreq))
        textLst.append("Nyquist frequency [Hz] = {}".format(self.nyquist))
        textLst.append("Number of time samples = {}".format(self.windowSize))
        textLst.append("Number of frequency samples = {}".format(self.dataSize))
        textLst.append("Number of channels = {}".format(self.numChans))
        textLst.append("Channels = {}".format(self.chans))
        textLst.append("Start time = {}".format(self.startTime))
        textLst.append("Stop time = {}".format(self.stopTime))
        if len(self.comments) == 0:
            textLst.append("No comments")
        else:
            textLst.append("Comments")
            for comment in self.comments:
                textLst.append("\t{}".format(comment))
        return textLst




import os
import glob
from copy import deepcopy
from datetime import datetime, timedelta
import numpy as np
from typing import List, Dict, Tuple, Any, Union

# import from package
from resistics.ioHandlers.ioHandler import IOHandler
from resistics.dataObjects.timeData import TimeData
from resistics.utilities.utilsPrint import blockPrint
from resistics.utilities.utilsChecks import isElectric, isMagnetic
from resistics.utilities.utilsClean import removeNansSingle, removeZerosSingle


class DataReader(IOHandler):
    """Base class for data readers

    DataReader is the base class for the time data readers. It implements much of the non-format dependent methods.

    The DataReaders for the different formats are meant to give a consistent output. Conventions used include:

    - The start time is the time of the first sample
    - The end time is the time of the last sample

    Physical data is always returned in:

    - Electrical channels in mV/km
    - Magnetic channels in mV
    - To get magnetic fields in nT, calibration needs to be performed    

    Attributes
    ----------
    dataPath : str
        Path to the data folder.
    headers : Dict
        Dictionary mapping header words to values
    chans : List[str]
        List of channels
    numChannels : int
        Number of channels
    chanHeaders : List	
        Headers specific to channels
    chanMap : Dict
        Map from channel name to index for chanHeaders
    comments : List[str]
        List of comments associated with data
    headerF	: List[str]
        List of header files (with extension .hdr)
    dataF : List[str]
        List of data files (with extension .npy)
    dataByteOffset : int
        Number of bytes to offset before reading
    dataByteSize : int
        Byte size of a single data value

    Methods
    -------
    __init__(dataPath)
        Initialise with path to the data directory
    setParameters()
        Set parameters specific to a data format
    checkFiles()
        Check to see header and data files found in data directory
    getComments()
        Get a deepcopy of the comments
    getHeaders()
        Get copy of header dictionary
    getChannels()
        Get a list of the channels
    getNumChannels()   
        Get number of channels
    getSampleFreq(integer = False)
        Get sampling frequency in Hz as float or optionally as integer
    getSampleRate()
        Get sample rate in s
    getNumSamples()
        Get the number of samples
    getStartDatetime()
        Get the data start time as a datetime object
    getStopDatetime()
        Get the data stop time as a datetime object
    getGain1() 
        Get gain stage 1 for each channel in an array
    getGain2()
        Get gain stage 2 for each channel in an array
    getChanMap()
        Get the channel map
    getChanHeader(chan, header)
        Get a channel header
    getChanType(chan)
        Get channel type (electric, magnetic)
    getChanGain1(chan)
        Get gain stage 1 for chan
    getChanGain2(chan)
        Get gain stage 2 for chan
    getChanSamples(chan)
        Get number of samples in channel
    getChanSampleFreq(chan)
        Get sampling frequency for a channel
    getChanLSB(chan)
        Get the channel least significant bit
    getChanScalingApplied(chan)
        Get a bool flag designating whether LSB has been applied
    getChanDataFile(chan)
        Get the channel data file
    getChanDx(chan)
        Get the distance between x electrodes
    getChanDy(chan)
        Get the distance between y electrodes
    getChanDz(chan)
        Get the distance between z electrodes
    getChanSensor(chan)
        Get sensor value of a channel
    getChanSerial(chan)
        Get serial value of a channel
    getChanChopper(chan)
        Get chopper value for a channel
    getSensors(chans)
        Get sensor values for a list of channels
    getSerials(chans)
        Get serial values for a list of channels
    getChoppers(chans)
        Get chopper values for a list of channels
    setHeader(headerName, headerVal) 
        Set a header value
    setChanHeader(chan, headerName, headerVal)
        Set a chan header value
    getUnscaledSamples(**kwargs)
        Get raw, unscaled data
    getUnscaledData(startTime, endTime, **kwargs)
        Get raw, unscaled data for a date range
    getPhysicalSamples(**kwargs)
        Get data in physical units
    getPhysicalData(startTime, endTime, **kwargs)
        Get data in physical units for a date range
    parseGetDataKeywords(keywords)
        Parse get data keywords
    time2sample(timeStart, timeEnd)
        Convert dates to samples
    sample2time(sampleStart, sampleEnd)
        Convert samples to datetimes
    getDataTimes(timeStart, timeEnd)
        Check data times and return checked and corrected ones
    readHeader()
        Read header data (implemented in child classes)
    formatHeaderData()
        Format the header data
    intHeaders()
        Return a list of headers to be formatted as int    
    floatHeaders()
        Return a list of headers to be formatted as float
    boolHeaders()
        Return a list of headers to be formatted as bool
    prepare()
        Prepare class information after reading header files
    checkChan(chan)
        Check channel exists in data
    printList()
        Class information as list of strings
    printCommentsList()
        Comment information as list of strings
    printComments()
        Print comments to terminal
    """

    def __init__(self, dataPath: str) -> None:
        """Initialise with path to data directory

        Parameters
        ----------
        dataPath : str
            Path to data directory
        """

        self.dataPath: str = dataPath
        self.headers: Dict = {}
        self.chans = []
        self.numChannels: int = 0
        self.chanHeaders: List = []
        self.chanMap: Dict = {}
        self.comments: List[str] = []
        # get a list of the xml files in the folder
        self.setParameters()
        if not self.checkFiles():
            self.printError("No header or data files found", quitRun=True)
        self.readHeader()
        self.formatHeaderData()
        self.prepare()

    def setParameters(self) -> None:
        """Set data reader parameters

        This will vary for the different data formats. By default, setup for the internal data format.
        """

        self.headerF = glob.glob(os.path.join(self.dataPath, "*.hdr"))
        self.dataF = glob.glob(os.path.join(self.dataPath, "*.dat"))
        self.dataByteOffset = 0
        self.dataByteSize = 4

    def checkFiles(self) -> bool:
        """Check to make sure data files found"""

        check = True
        if len(self.headerF) == 0:
            check = check and False
            self.printWarning("No header files found")
        if len(self.dataF) == 0:
            check = check and False
            self.printWarning("No data files found")
        return check

    def getComments(self) -> List[str]:
        """Get a deepcopy of the comments
        
        Returns
        -------
        List[str]
            Dataset comments as a list of strings
        """

        return deepcopy(self.comments)

    def getHeaders(self) -> Dict:
        """Get the data headers

        Returns
        -------
        Dict
            Data headers
        """

        return deepcopy(self.headers)

    def getChannels(self) -> List[str]:
        """Get channels in data

        Returns
        -------
        List[str]
            Data channels
        """

        return deepcopy(self.chans)

    def getNumChannels(self) -> int:
        """Get number of channels in the data

        Returns
        -------
        int
            Number of channels
        """

        return self.headers["meas_channels"]

    def getSampleFreq(self, integer: bool = False):
        """Get data sampling frequency in Hz

        Returns
        -------
        float, int
            Sampling frequency
        """

        if integer:
            return int(self.headers["sample_freq"])
        return self.headers["sample_freq"]

    def getSampleRate(self) -> float:
        """Get data sampling rate in  s

        Returns
        -------
        float
            Sampling rate in s
        """

        return 1.0 / self.getSampleFreq()

    def getNumSamples(self) -> int:
        """Get number of samples

        Returns
        -------
        int
            Number of samples
        """

        return self.headers["num_samples"]

    def getStartDatetime(self) -> datetime:
        """Get datetime of first sample

        Returns
        -------
        datetime
            Date and time of first sample
        """

        return deepcopy(self.datetimeStart)

    def getStopDatetime(self) -> datetime:
        """Get datetime of last sample

        Returns
        -------
        datetime
            Date and time of last sample
        """

        return deepcopy(self.datetimeStop)

    def getGain1(self) -> np.ndarray:
        """Get value of gain 1

        Returns
        -------
        np.ndarray
            Array of gains for channels
        """

        gain1 = np.zeros(shape=(self.numChannels), dtype=bool)
        for iChan in range(0, self.numChannels):
            gain1[iChan] = self.getChanGain1(iChan)
        return gain1

    def getGain2(self) -> np.ndarray:
        """Get value of gain 2

        Returns
        -------
        np.ndarray
            Array of gains for channels
        """

        gain2 = np.zeros(shape=(self.numChannels), dtype=bool)
        for iChan in range(0, self.numChannels):
            gain2[iChan] = self.getChanGain2(iChan)
        return gain2

    def getChanHeaders(self, chans: List[str] = []) -> Tuple[List, Dict]:
        """Get channel headers

        Parameters
        ----------
        chans: List[str], optional
            List of channels for which chan headers are wanted
        Returns
        -------
        chanHeaders : List
            List of channel headers
        chanMap : Dict
            Map from channel to index for the chanHeaders
        """

        if len(chans) == 0:
            return deepcopy(self.chanHeaders), deepcopy(self.chanMap)

        chanHeaders = []
        chanMap = {}
        for idx, c in enumerate(chans):
            chanIdx = self.getChanMap()[c]
            chanHeaders.append(self.chanHeaders[chanIdx])
            chanMap[c] = idx
        return chanHeaders, chanMap

    def getChanMap(self):
        return self.chanMap

    def getChanHeader(self, chan, header):
        self.checkChan(chan)
        iChan = self.chanMap[chan]
        return self.chanHeaders[iChan][header]

    def getChanType(self, chan) -> str:
        """Get the channel type (electric or magnetic)

        Returns
        -------
        str
            String of channel type
        """

        return self.getChanHeader(chan, "channel_type")

    def getChanGain1(self, chan) -> int:
        """Get channel gain 1 

        Returns
        -------
        int
            Channel gain 1
        """

        return self.getChanHeader(chan, "gain_stage1")

    def getChanGain2(self, chan) -> int:
        """Get channel gain 2

        Returns
        -------
        int
            Channel gain 2
        """

        return self.getChanHeader(chan, "gain_stage2")

    def getChanSamples(self, chan) -> int:
        """Get channel number of samples

        Returns
        -------
        int
            Channel number of samples
        """

        return self.getChanHeader(chan, "num_samples")

    def getChanSampleFreq(self, chan) -> float:
        """Get channel sampling frequency 

        Returns
        -------
        float
            Sampling frequency in Hz
        """

        return self.getChanHeader(chan, "sample_freq")

    def getChanLSB(self, chan):
        """Get channel least significant bit 

        Returns
        -------
        float
            Channel least significant bit
        """

        return self.getChanHeader(chan, "ts_lsb")

    def getChanScalingApplied(self, chan) -> bool:
        """A flag to mark whether a channel has the lsb applied

        Returns
        -------
        bool
            Flag to designate whether channel lsb applied
        """

        return self.getChanHeader(chan, "scaling_applied")

    def getChanDataFile(self, chan) -> str:
        """Get the data file for the channel 

        Returns
        -------
        str
            Data file for the channel
        """

        return self.getChanHeader(chan, "ats_data_file")

    def getChanDx(self, chan):
        """Get the electric channel spacing in the x direction 

        Returns
        -------
        float
            Electric channel spacing in x direction in metres
        """

        x1 = np.absolute(self.getChanHeader(chan, "pos_x1"))
        x2 = np.absolute(self.getChanHeader(chan, "pos_x2"))
        return x2 + x1

    def getChanDy(self, chan):
        """Get the electric channel spacing in the y direction 

        Returns
        -------
        float
            Electric channel spacing in y direction in metres
        """

        y1 = np.absolute(self.getChanHeader(chan, "pos_y1"))
        y2 = np.absolute(self.getChanHeader(chan, "pos_y2"))
        return y2 + y1

    def getChanDz(self, chan):
        """Get the electric channel spacing in the z direction 

        Returns
        -------
        float
            Electric channel spacing in z direction in metres
        """

        z1 = np.absolute(self.getChanHeader(chan, "pos_z1"))
        z2 = np.absolute(self.getChanHeader(chan, "pos_z2"))
        return z2 + z1

    def getChanSensor(self, chan) -> str:
        """Get channel sensor type

        Returns
        -------
        str
            Channel sensor type
        """

        return self.getChanHeader(chan, "sensor_type")

    def getChanSerial(self, chan) -> str:
        """Get channel serial number

        Returns
        -------
        str
            Channel serial number
        """

        return self.getChanHeader(chan, "sensor_sernum")

    def getChanChopper(self, chan) -> bool:
        """Get channel chopper
        
        The chopper is an amplifier present in some instruments. There might be different calibration files for chopper on or off.

        Returns
        -------
        bool
            Flag designating whether chopper is on or off
        """

        echopper = self.getChanHeader(chan, "echopper")
        hchopper = self.getChanHeader(chan, "hchopper")
        # return true if the chopper amplifier was on
        if isElectric(chan) and echopper:
            return True
        if isMagnetic(chan) and hchopper:
            return True
        return False

    def getSensors(self, chans: List[str]) -> Dict[str, str]:
        """Get sensors for multiple chans

        Returns
        -------
        Dict[str, str]
            Dictionary with channels as keys and the sensor types as values 
        """

        sensors = {}
        for chan in chans:
            sensors[chan] = self.getChanSensor(chan)
        return sensors

    def getSerials(self, chans: List[str]) -> Dict[str, str]:
        """Get serials for multiple chans

        Returns
        -------
        Dict[str, str]
            Dictionary with channels as keys and the serials numbers as values 
        """

        serials = {}
        for chan in chans:
            serials[chan] = self.getChanSerial(chan)
        return serials

    def getChoppers(self, chans: List[str]) -> Dict[str, bool]:
        """Get choppers for multiple chans

        Returns
        -------
        Dict[str, str]
            Dictionary with channels as keys and the serials numbers as values 
        """

        choppers = {}
        for chan in chans:
            choppers[chan] = self.getChanChopper(chan)
        return choppers

    def setHeader(self, headerName: str, headerVal: Any) -> None:
        """Set a header value

        Parameters
        ----------
        headerName : str
            The name of the header to set
        headerVal : Any
            Header value
        """

        self.headerName = headerVal

    def setChanHeader(self, chan: str, headerName: str, headerVal: Any) -> None:
        """Set a channel header value

        Parameters
        ----------
        channel : str
            The channel
        headerName : str
            The name of the header to set
        headerVal : Any
            Header value
        """

        chanIndex = self.chanMap[chan]
        self.chanHeaders[chanIndex][headerName] = headerVal

    def getUnscaledSamples(self, **kwargs) -> TimeData:
        """Get raw data from data file

        Depending on the data format, this could be raw counts or in some physical unit. The method implemented in the base DataReader can read from ATS and internal files. SPAM and Phoenix data readers have their own implementations.

        The raw data units for ATS and internal data formats are as follows: 

        - ATS data format has raw data in counts.
        - The raw data unit of the internal format is dependent on what happened to the data before writing it out in the internal format. If the channel header scaling_applied is set to True, no scaling happens in either getUnscaledSamples or getPhysicalSamples. However, if the channel header scaling_applied is set to False, the internal format data will be treated like ATS data, meaning raw data in counts.
        
        Parameters
        ----------
        chans : List[str], optional
            List of channels to return if not all are required
        startSample : int, optional
            First sample to return
        endSample : int, optional
            Last sample to return

        Returns
        -------
        TimeData
            Time data object
        """

        # initialise chans, startSample and endSample with the whole dataset
        options = self.parseGetDataKeywords(kwargs)
        # get samples - this is inclusive
        dSamples = options["endSample"] - options["startSample"] + 1

        # loop through chans and get data
        data = {}
        for chan in options["chans"]:
            # check to make sure channel exists
            self.checkChan(chan)
            # get data file
            dFile = os.path.join(self.dataPath, self.getChanDataFile(chan))
            # get the data
            byteOff = self.dataByteOffset + options["startSample"] * self.dataByteSize
            # now check if lsb applied or not and read data as float32 or int32 accordingly
            if self.getChanScalingApplied(chan):
                data[chan] = np.memmap(
                    dFile, dtype="float32", mode="r", offset=byteOff, shape=(dSamples)
                )
            else:
                data[chan] = np.memmap(
                    dFile, dtype="int32", mode="r", offset=byteOff, shape=(dSamples)
                )

        # get data start and stop time
        startTime, stopTime = self.sample2time(
            options["startSample"], options["endSample"]
        )
        # dataset comments
        comments = []
        comments.append(
            "Unscaled data {} to {} read in from measurement {}, samples {} to {}".format(
                startTime,
                stopTime,
                self.dataPath,
                options["startSample"],
                options["endSample"],
            )
        )
        comments.append("Sampling frequency {}".format(self.getSampleFreq()))
        if len(self.comments) > 0:
            comments = self.comments + comments
        return TimeData(
            sampleFreq=self.getSampleFreq(),
            startTime=startTime,
            stopTime=stopTime,
            data=data,
            comments=comments,
        )

    def getUnscaledData(self, startTime, endTime, **kwargs) -> TimeData:
        """Get raw data from data file between a start and end date

        Calculates the start and end sample given the data range and returns getUnscaledSamples for that sample range.

        Parameters
        ----------
        startTime : datetime
            Start time of data to read
        endTime : datetime
            End time of data to read       
        chans : List[str], optional
            List of channels to return if not all are required

        Returns
        -------
        TimeData
            Time data object
        """

        options = self.parseGetDataKeywords(kwargs)
        startSample, endSample = self.time2sample(startTime, endTime)
        return self.getUnscaledSamples(
            chans=options["chans"], startSample=startSample, endSample=endSample
        )

    def getPhysicalSamples(self, **kwargs):
        """Get data scaled to physical values
        
        Depending on the data format, the scalings required to convert to field physical units is different. The method in the base DataReader class covers ATS and internal file format.

        resistics will always provide physical samples in field units. That means

        - Electrical channels in mV/km
        - Magnetic channels in mV
        - To get magnetic fields in nT, calibration needs to be performed
        
        If the channel header scaling_applied is set to True, no scaling of the unscaled data is done. This is to cover the internal data format where all scalings may already have been applied.

        Notes
        -----
        The raw data units for ATS data are in counts. To get data in field units, ATS data is first multipled by the least significat bit (lsb) defined in the header files,

        .. code-block:: text  
        
            data = data * leastSignificantBit,
        
        giving data in mV. The lsb includes the gain removal, so no separate gain removal needs to be performed.
        
        For electrical channels, there is additional step of dividing by the electrode spacing, which is provided in metres. The extra factor of a 1000 is to convert this to km to give mV/km for electric channels
        
        .. code-block:: text  
            
            data = (1000 * data)/electrodeSpacing
        
        Finally, to get magnetic channels in nT, the magnetic channels need to be calibrated.

        Parameters
        ----------
        chans : List[str]
            List of channels to return if not all are required
        startSample : int
            First sample to return
        endSample : int
            Last sample to return
        remaverage : bool
            Remove average from the data
        remzeros : bool
            Remove zeroes from the data
        remnans: bool
            Remove NanNs from the data

        Returns
        -------
        TimeData
            Time data object
        """

        options = self.parseGetDataKeywords(kwargs)
        timeData = self.getUnscaledSamples(
            chans=options["chans"],
            startSample=options["startSample"],
            endSample=options["endSample"],
        )
        # multiply each chan by least significant bit of chan
        for chan in options["chans"]:
            if not self.getChanScalingApplied(chan):
                # apply LSB to give data in mV
                timeData.data[chan] = timeData.data[chan] * self.getChanLSB(chan)
                timeData.addComment(
                    "Scaling channel {} with scalar {} to give mV".format(
                        chan, self.getChanLSB(chan)
                    )
                )

                # divide by the distance - this should only be for the electric channels
                # again, this might already be applied
                if chan == "Ex":
                    # multiply by 1000/self.getChanDx same as dividing by dist in km
                    timeData.data[chan] = (
                        1000 * timeData.data[chan] / self.getChanDx(chan)
                    )
                    timeData.addComment(
                        "Dividing channel {} by electrode distance {} km to give mV/km".format(
                            chan, self.getChanDx(chan) / 1000.0
                        )
                    )
                if chan == "Ey":
                    # multiply by 1000/self.getChanDy same as dividing by dist in km
                    timeData.data[chan] = (
                        1000 * timeData.data[chan] / self.getChanDy(chan)
                    )
                    timeData.addComment(
                        "Dividing channel {} by electrode distance {} km to give mV/km".format(
                            chan, self.getChanDy(chan) / 1000.0
                        )
                    )

            # if remove zeros - False by default
            if options["remzeros"]:
                timeData.data[chan] = removeZerosSingle(timeData.data[chan])
            # if remove nans - False by default
            if options["remnans"]:
                timeData.data[chan] = removeNansSingle(timeData.data[chan])
            # remove the average from the data - True by default
            # do this after all scaling and removing nans and zeros
            if options["remaverage"]:
                timeData.data[chan] = timeData.data[chan] - np.average(
                    timeData.data[chan]
                )

        timeData.addComment(
            "Remove zeros: {}, remove nans: {}, remove average: {}".format(
                options["remzeros"], options["remnans"], options["remaverage"]
            )
        )
        return timeData

    def getPhysicalData(self, startTime, endTime, **kwargs):
        """Get physical data from data file between a start and end data

        Calculates the start and end sample given the data range and returns getPhysicalSamples for that sample range.
        
        Parameters
        ----------
        startTime : datetime
            Start time of data to read
        endTime : datetime
            End time of data to read       
        chans : List[str]
            List of channels to return if not all are required
        remaverage : bool
            Remove average from the data
        remzeros : bool
            Remove zeroes from the data
        remnans: bool
            Remove NanNs from the data

        Returns
        -------
        TimeData
            Time data object
        """

        options = self.parseGetDataKeywords(kwargs)
        startSample, endSample = self.time2sample(startTime, endTime)
        return self.getPhysicalSamples(
            chans=options["chans"], startSample=startSample, endSample=endSample
        )

    def parseGetDataKeywords(self, keywords) -> Dict:
        """Parse the get data keywords
        
        Parameters
        ----------
        keywords : Dict
            The keywords passed to get data methods

        Returns
        -------
        Dict
            A dictionary of parsed keywords with defaults where nothing is provided by the user
        """

        # defaults
        options = {}
        options["chans"] = self.getChannels()
        options["startSample"] = 0
        options["endSample"] = self.getNumSamples() - 1
        options["startTime"] = self.getStartDatetime()
        options["endTime"] = self.getStopDatetime()
        options["remaverage"] = True
        options["remzeros"] = False
        options["remnans"] = False
        # now take the options from the keywords
        for w in options:
            if w in keywords:
                options[w] = keywords[w]
        # do some checks
        if options["endSample"] >= self.getNumSamples():
            options["endSample"] = self.getNumSamples() - 1
            self.printWarning(
                "End sample greater than number of samples. Adjusted to {:d}".format(
                    options["endSample"]
                )
            )
        if options["startSample"] >= options["endSample"]:
            options["startSample"] = (
                options["endSample"] - 100
            )  # let's give 100 samples
            self.printWarning(
                "Start sample greater than end sample. Adjusted to {:d}".format(
                    options["startSample"]
                )
            )
        if options["startSample"] < 0:
            options["startSample"] = 0
            self.printWarning(
                "Start sample < 0. Adjusted to {:d}".format(options["startSample"])
            )
        return options

    def time2sample(
        self, timeStart: Union[str, datetime], timeEnd: Union[str, datetime]
    ) -> Tuple[int, int]:
        """Converts a start and end time to start and end samples
        
        Note: The first sample is zero

        Parameters
        ----------
        timeStart : datetime, str
            Start time of date range
        timeEnd : datetime, str
            End time of date range       

        Returns
        -------
        sampleStart : int
            The correspoding start sample for timeStart
        sampleEnd : int
            The corresponding end sample for timeEnd
        """

        # if timeStart and timeEnd are strings, then convert them to datetime objects
        if isinstance(timeStart, str):
            timeStart = datetime.strptime(timeStart, "%Y-%m-%d %H:%M:%S")
        if isinstance(timeEnd, str):
            timeEnd = datetime.strptime(timeEnd, "%Y-%m-%d %H:%M:%S")
        # check to see times within range
        timeStart, timeEnd = self.getDataTimes(timeStart, timeEnd)
        # start sample
        deltaStart = timeStart - self.getStartDatetime()
        sampleStart = deltaStart.total_seconds() * self.getSampleFreq()
        sampleStart = int(
            round(sampleStart)
        )  # this will hopefully deal with fractional sampling
        # end sample
        deltaEnd = timeEnd - timeStart
        deltaSamples = deltaEnd.total_seconds() / self.getSampleRate()
        deltaSamples = int(round(deltaSamples))
        sampleEnd = sampleStart + deltaSamples
        # return samples
        return sampleStart, sampleEnd

    def sample2time(
        self, sampleStart: int, sampleEnd: int
    ) -> Tuple[datetime, datetime]:
        """Converts a start and end sample to start and end times
        
        Note: The first sample is zero

        Parameters
        ----------
        sampleStart : int
            The starting sample for the sample range
        sampleEnd : int
            The ending sample for the sample range   

        Returns
        -------
        timeStart : datetime, str
            Corresponding start time of date range
        timeEnd : datetime, str
            Corresponding end time of date range            
        """

        # convert samples to some data format
        deltaStart = timedelta(seconds=self.getSampleRate() * sampleStart)
        # delta end is inclusive
        deltaEnd = timedelta(seconds=self.getSampleRate() * (sampleEnd - sampleStart))
        timeStart = self.getStartDatetime() + deltaStart
        timeEnd = timeStart + deltaEnd
        return timeStart, timeEnd

    def getDataTimes(self, timeStart, timeEnd):
        """Checks and converts a date range to make sure it's within data start and end
        
        Parameters
        ----------
        timeStart : datetime, str
            Start time of date range
        timeEnd : datetime, str
            End time of date range     

        Returns
        -------
        timeStart : datetime, str
            Checked and corrected start time of data range
        timeEnd : datetime, str
            Checked and corrected end time of data range        
        """

        deltaStart = timeStart - self.getStartDatetime()
        deltaEnd = self.getStopDatetime() - timeEnd
        if deltaStart.total_seconds() < 0:
            self.printText(
                "Date {} before start of recording. Start date adjusted to {}".format(
                    timeStart, self.getStartDatetime()
                )
            )
            timeStart = self.getStartDatetime()
        if deltaEnd.total_seconds() < 0:
            self.printText(
                "Date {} after end of recording. Stop date adjusted to {}".format(
                    timeEnd, self.getStopDatetime()
                )
            )
            timeEnd = self.getStopDatetime()
        return timeStart, timeEnd

    def readHeader(self) -> None:
        """Function to read header data and populate reader information
        
        This is implemented in child classes as all header formats are different
        """

        raise NotImplementedError(
            "Read header not implemented in parent class. Only child classes should ever be instantiated."
        )

    def formatHeaderData(self) -> None:
        """Format header data to the correct formats"""

        # do the int formatting
        intGlobal, intChan = self.intHeaders()
        floatGlobal, floatChan = self.floatHeaders()
        boolGlobal, boolChan = self.boolHeaders()
        # deal with the global headers
        for h in intGlobal:
            self.headers[h] = int(self.headers[h])
        for h in floatGlobal:
            self.headers[h] = float(self.headers[h])
        for h in boolGlobal:
            self.headers[h] = bool(self.headers[h])
        # deal with the channel headers
        numChans = len(self.chanHeaders)
        for iChan in range(0, numChans):
            for cH in self.chanHeaders[iChan]:
                if cH in intChan:
                    self.chanHeaders[iChan][cH] = int(self.chanHeaders[iChan][cH])
                if cH in floatChan:
                    self.chanHeaders[iChan][cH] = float(self.chanHeaders[iChan][cH])
                if cH in boolChan:
                    if isinstance(
                        self.chanHeaders[iChan][cH], str
                    ):  # if string, parse true and false
                        if self.chanHeaders[iChan][cH] == "True":
                            self.chanHeaders[iChan][cH] = True
                        else:
                            self.chanHeaders[iChan][cH] = False
                    else:  # a bool or an integer
                        self.chanHeaders[iChan][cH] = bool(self.chanHeaders[iChan][cH])

    def intHeaders(self) -> Tuple[List[str], List[str]]:
        """List of headers which are expected to have integer values
        
        Returns
        -------
        intGlobal : List[str]
            List of global integer headers
        intChan : List[str]
            List of chan integer headers
        """

        intGlobal = ["meas_channels"]
        intChan = [
            "gain_stage1",
            "gain_stage2",
            "hchopper",
            "echopper",
            "num_samples",
            "sensor_sernum",
        ]
        return intGlobal, intChan

    def floatHeaders(self) -> Tuple[List[str], List[str]]:
        """List of headers which are expected to have boolean values
        
        Returns
        -------
        floatGlobal : List[str]
            List of global float headers
        floatChan : List[str]
            List of chan float headers
        """

        floatGlobal = ["sample_freq"]
        floatChan = [
            "sample_freq",
            "ts_lsb",
            "pos_x1",
            "pos_x2",
            "pos_y1",
            "pos_y2",
            "pos_z1",
            "pos_z2",
        ]
        return floatGlobal, floatChan

    def boolHeaders(self) -> Tuple[List[str], List[str]]:
        """List of headers which are expected to have boolean values
        
        Returns
        -------
        boolGlobal : List[str]
            List of global boolean headers
        boolChan : List[str]
            List of chan boolean headers
        """

        boolGlobal = []
        boolChan = ["scaling_applied"]
        return boolGlobal, boolChan

    def prepare(self) -> None:
        """Set some intial values
        
        This method does some checks and prepares some of the storage for the channels.                 
        
        Notes
        -----
        The end time of the data will be checked. Different data formats record an end time after the last sample. For example, ATS end time appears to be one sample after the number of samples. The end time is checked by using the number of samples and the sampling frequency. 
        
        The internal convention is that the start and end times should reflect the times of the first and last sample
        """

        # create the type - index map
        self.chans = []
        self.chanMap = {}
        for iChan in range(0, self.getNumChannels()):
            chanType = self.chanHeaders[iChan]["channel_type"]
            self.chanMap[chanType] = iChan
            self.chans.append(chanType)

        # check the number of samples of each channel
        numSamples = []
        for c in self.chans:
            numSamples.append(self.getChanSamples(c))

        self.headers["num_samples"] = min(numSamples)
        for c, n in zip(self.chans, numSamples):
            if n != self.getNumSamples():
                self.printWarning("Not all channels have the same number of samples")
                self.printWarning(
                    "{} has {:d} samples more than the minimum".format(
                        c, n - self.getNumSamples()
                    )
                )

        # create datetime objects
        datetimeStart = "{} {}".format(
            self.headers["start_date"], self.headers["start_time"]
        )
        datetimeStop = "{} {}".format(
            self.headers["stop_date"], self.headers["stop_time"]
        )
        self.datetimeStart = datetime.strptime(datetimeStart, "%Y-%m-%d %H:%M:%S.%f")
        self.datetimeStop = datetime.strptime(datetimeStop, "%Y-%m-%d %H:%M:%S.%f")

        # check the stop time
        startTime, endTime = self.sample2time(0, self.getNumSamples() - 1)
        if endTime != self.datetimeStop:
            self.datetimeStop = endTime
            self.headers["stop_date"] = self.datetimeStop.strftime("%Y-%m-%d")
            self.headers["stop_time"] = self.datetimeStop.strftime("%H:%M:%S.%f")
            for idx, chan in enumerate(self.chanHeaders):
                self.chanHeaders[idx]["stop_date"] = self.datetimeStop.strftime(
                    "%Y-%m-%d"
                )
                self.chanHeaders[idx]["stop_time"] = self.datetimeStop.strftime(
                    "%H:%M:%S.%f"
                )

    def checkChan(self, chan: str) -> None:
        """Check channel exists in data

        Parameters
        ----------
        chan : str
            Channel to check
        """

        if chan not in self.chans:
            self.printError(
                "Error - Channel {} does not exist".format(chan), quitRun=True
            )

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Data Path = {}".format(self.dataPath))
        textLst.append("Global Headers")
        textLst.append("{}".format(self.headers))
        textLst.append("Channels found:")
        textLst.append("{}".format(self.chans))
        textLst.append("Channel Map")
        textLst.append("{}".format(self.chanMap))
        textLst.append("Channel Headers")
        for c in self.chans:
            textLst.append(c)
            textLst.append("{}".format(self.chanHeaders[self.chanMap[c]]))
        textLst.append(
            "Note: Field units used. Physical data has units mV/km for electric fields and mV for magnetic fields"
        )
        textLst.append("Note: To get magnetic field in nT, please calibrate")
        return textLst

    def printCommentsList(self) -> List[str]:
        """Dataset comments as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst: List[str] = []
        textLst.append("Dataset Comments")
        if len(self.comments) == 0:
            textLst.append("No comments")
        else:
            for idx, comment in enumerate(self.comments):
                textLst.append("{:02d} : {}".format(idx, comment))
        return textLst

    def printComments(self) -> None:
        """Print out dataset comments"""

        blockPrint(
            "{} Comments".format(self.__class__.__name__), self.printCommentsList()
        )




import math
import numpy as np
import scipy.signal as signal
from typing import List

# import from package
from resistics.utilities.utilsMath import intdiv


def smooth1d(x: np.ndarray, winLen: int = 11, window: str = "hanning"):
    """Smooth in 1 dimension

    Parameters
    ----------
    x : np.ndarray
        The data array to be smoothed
    winLen : int
        The number of samples to smooth across
    window : str
        The window function
	
    Notes
    ----- 
    The data is padded before the smoothing to ensure that the output data is the same size as the input. 

    todo:
    The window parameter could be the window itself if an array instead of a string
	"""

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")
    if x.size < winLen:
        raise ValueError("Input vector needs to be bigger than window size.")
    if winLen < 3:
        return x

    s = np.pad(x, (winLen, winLen), mode="edge")
    if window == "flat":  # moving average
        w = np.ones(winLen, "d")
    else:
        w = eval("signal." + window + "(winLen)")

    # assume winLen is odd
    off = winLen + intdiv((winLen - 1), 2)
    if winLen % 2 == 0:  # check if even and recalc off
        off = winLen + intdiv(winLen, 2)

    y = np.convolve(s, w / w.sum(), mode="full")
    return y[off : off + x.size]


def smooth2d(x: np.ndarray, winLen: List[int] = [5, 5], window: str = "hanning"):
    """Smooth in 2-D

    This smooths in a window and across windows too

    Parameters
    ----------
    x : np.ndarray
        A 2-D array to be smoothed
    winLen : List[int]
        A two element list with winLen[0] being smoothing across windows and winLen[1] the smoothing within a window
    window : str
        The window function to use
    """

    kernel = np.outer(signal.hanning(winLen[0], 8), signal.gaussian(winLen[1], 8))
    # pad to help the boundaries
    padded = np.pad(x, ((winLen[0], winLen[0]), (winLen[1], winLen[1])), mode="edge")
    # 2d smoothing
    blurred = signal.fftconvolve(padded, kernel, mode="same")
    return blurred[
        winLen[0] : winLen[0] + x.shape[0], winLen[1] : winLen[1] + x.shape[1]
    ]



import os
from datetime import datetime, timedelta
import numpy as np
import matplotlib.pyplot as plt
from typing import List, Union

# import from package
from resistics.dataObjects.dataObject import DataObject
from resistics.utilities.utilsPlotter import getViewFonts


class CalibrationData(DataObject):
    """Class for holding calibration data

    Calibration data should be given in the frequency domain and has a magnitude and phase component (in radians). Calibration data is the impulse response for an instrument or sensor and is usually deconvolved (division in frequency domain) from the time data.
    
    Notes
    -----
    Calibration data for magnetic channels is given in mV/nT. Because this is deconvolved from magnetic time data, which is in mV, the resultant magnetic time data is in nT. 

    Attributes
    ----------
    filename : str
        The filename the calibration data was read from
    numSamples : int
        The number of samples in the calibration data
    freqs : np.ndarray
        The frequency points where calibration data is defined
    magnitude : np.ndarray
        The magnitude data
    magnitudeUnit : str
        The magnitude unit, defaulted to mV/nT
    phase : np.ndarray
        The phase data in radians
    phaseUnit : str
        The phase unit, defaulted to radians
    chopper : bool
        Boolean flag to note whether chopper is on or not

    Methods
    -------
    __init__(kwargs)
        Initialise the calibration data
    view(kwargs)
        View the calibration data 
    printList()
        Class status returned as list of strings          
    """

    def __init__(
        self,
        filename: str,
        freqs: np.ndarray,
        magnitude: np.ndarray,
        phase: np.ndarray,
        staticGain: float = 1,
        chopper: bool = False,
        serial: int = 1,
        sensor: str = "",
    ) -> None:
        """Initialise and set object parameters

        Parameters
        ----------
        filename : str
            The filename the calibration was read in from
        freqs : np.ndarray
            Array of frequencies for which the impulse response is defined
        magnitude : np.ndarray
            Magnitude of impulse response
        phase : np.ndarray
            Phase of impulse reponse in radians 
        staticGain : float, optional
            The static gain to applied to the calibration data
        chopper : bool, optional
            Boolean flag for chopper on or off
        serial : int
            The serial number of the sensor or instrument which requires calibration
        sensor : str
            The sensor name of the sensor or instrument which requires calibration
        """

        self.filename: str = filename
        self.freqs: np.ndarray = freqs
        self.magnitude: np.ndarray = magnitude
        self.magnitudeUnit = "mV/nT"
        self.phase: np.ndarray = phase
        self.phaseUnit = "radians"
        self.numSamples: int = len(freqs)
        self.staticGain: float = staticGain
        self.chopper: bool = chopper
        self.serial: int = serial
        self.sensor: bool = sensor

    def view(self, **kwargs) -> plt.figure:
        """Plot of the calibration function

        Parameters
        ----------
        staticgain : bool, optional
            Boolean flag for having static gain on, default is True
        degrees : bool, optional
            Plot phase in degreesm default is False
        fig : matplotlib.pyplot.figure, optional
            A figure object
        plotFonts : Dict, optional
            A dictionary of plot fonts
        label : str, optional
            Label for the plots
        xlim : List, optional
            Limits for the x axis
        ylim_mag : List, optional
            Limits for the magnitude y axis
        ylim_phase : List, optional
            Limits for the phase y axis
        legened : bool
            Boolean flag for adding a legend
        
        Returns
        -------
        plt.figure
            Matplotlib figure object            
        """

        if "fig" in kwargs:
            fig = plt.figure(kwargs["fig"].number)
        else:
            fig = plt.figure(figsize=(8, 8))
        plotFonts = kwargs["plotFonts"] if "plotFonts" in kwargs else getViewFonts()

        # static gain
        magnitude = self.magnitude
        if "staticgain" in kwargs and not kwargs["staticgain"]:
            magnitude = magnitude / self.staticGain
        # phase
        phaseUnit = self.phaseUnit
        phase = self.phase
        if "degrees" in kwargs and kwargs["degrees"]:
            phaseUnit = "degrees"
            phase = self.phase * (180 / np.pi)

        # plot magnitude
        plt.subplot(2, 1, 1)
        plt.title("Impulse response magnitude", fontsize=plotFonts["title"])
        lab = kwargs["label"] if "label" in kwargs else self.filename
        plt.loglog(self.freqs, magnitude, label=lab)
        if "xlim" in kwargs:
            plt.xlim(kwargs["xlim"])
        if "ylim_mag" in kwargs:
            plt.ylim(kwargs["ylim_mag"])
        plt.xlabel("Frequency [Hz]")
        plt.ylabel("Magnitude [{}]".format(self.magnitudeUnit))
        plt.grid(True)
        # legend
        if "legend" in kwargs and kwargs["legend"]:
            plt.legend(loc=2)

        # plot phase
        plt.subplot(2, 1, 2)
        plt.title("Impulse response phase", fontsize=plotFonts["title"])
        lab = kwargs["label"] if "label" in kwargs else self.filename
        plt.semilogx(self.freqs, phase, label=lab)
        if "xlim" in kwargs:
            plt.xlim(kwargs["xlim"])
        if "ylim_phase" in kwargs:
            plt.ylim(kwargs["ylim_phase"])
        plt.xlabel("Frequency [Hz]")
        plt.ylabel("Phase [{}]".format(phaseUnit))
        plt.grid(True)
        # legend
        if "legend" in kwargs and kwargs["legend"]:
            plt.legend(loc=3)

        # show if the figure is not in keywords
        if "fig" not in kwargs:
            plt.tight_layout(rect=[0, 0.02, 1, 0.96])
            plt.show()

        return fig

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Filename = {}".format(self.filename))
        textLst.append("Serial = {}".format(self.serial))
        textLst.append("Sensor = {}".format(self.sensor))
        textLst.append("Static gain = {:.2f}".format(self.staticGain))
        textLst.append("Chopper = {}".format(self.chopper))
        textLst.append("Number of frequency points = {:d}".format(self.numSamples))
        textLst.append("Calibration data:")
        textLst.append(
            "\t{:>17s}\t{:>10s}\t{:>10s}".format(
                "Frequency [Hz]", "Mag. [mv/nT]", "Phase [rad]"
            )
        )
        for ii in range(0, self.numSamples):
            textLst.append(
                "\t{:17.8f}\t{:10.2f}\t{:10.2f}".format(
                    self.freqs[ii], self.magnitude[ii], self.phase[ii]
                )
            )
        return textLst



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from datetime import datetime, timedelta
from typing import List, Dict, Set

# import from package
from resistics.dataObjects.dataObject import DataObject
from resistics.utilities.utilsPrint import arrayToString, blockPrint
from resistics.utilities.utilsFreq import getFrequencyArray
from resistics.utilities.utilsPlotter import getViewFonts


class MaskData(DataObject):
    """Class for holding information about window masking

    Attributes
    ----------
    maskName : str
        The name of the mask
    sampleFreq : float  
        The sampling frequency of the data in Hz
    numLevels : int
        Number of decimation levels
    evalFreq : List[float]
        The evaluation frequnecies
    constraints : Dict
        Constraints index by evaluation frequency, statistic and component
    insideOut : Dict
        Flag stating whether to mask values inside constraints (False) or outside constraints (True), indexed by evaluation frequency, statistic and component
    maskWindows : Dict
        Dictionary for storing the masked windows for each evaluation frequency
    stats : List
        Statistics to use in the masking

    Methods
    -------
    __init__(sampleFreq, numLevels, evalFreq, kwargs)
        Initialise maskData
    setStats(stats)
        Set the statistics used in the MaskData
    addConstraint(stat, constraint, insideOut)
        Add constraints for statistic stat for all evaluation frequencies
    addConstraintLevel(stat, constraint, level, insideOut)
        Add constraints for statistic stat for all evaluation frequencies for a single decimation level  
    addConstraintFreq(stat, constraint, level, eIdx, insideOut)
        Add a constraint for statistic stat for a particular evaluation frequency               
    getConstraintFreq(level, eIdx, stat = "")
        Get the constraints for a evaluation frequency provided by the level and the evaluation frequency index and optional request constraints only for a particular statistic
    getMaskWindowsFreq(level, eIdx)
        Get a set of the masked windows for an evaluation frequency identified by decimation level and evaluation frequency index
    prepareDicts()
        Prepare the storage dictionaries given a new set of statistics
    resetMaskWindows()
        Clear the masked windows
    view(level)
        View the masked windows at a particular decimation level
    printList()
        Class status returned as list of strings
    printConstraints()
        Print constraint information to the console
    printConstraintsList()
        Return constraint information as a list of strings
    """

    def __init__(self, sampleFreq: float, numLevels: int, evalFreq, **kwargs):
        """Initialise mask data 
    
        Parameters
        ----------
        sampleFreq : float
            Data sampling frequency in Hz
        numLevels : int
            Number of decimation levels
        evalFreq : np.ndarray
            2-D Matrix, numlevels * evaluations frequencies per level. Contents are evaluations frequencies
        kwargs : Dict
            Two optional arguments: "name" for statistic name and "stats" for the statistics to use
        """

        self.maskName: str = "mask"
        if "name" in kwargs:
            self.maskName = kwargs["name"]
        self.sampleFreq: float = sampleFreq
        self.numLevels: int = numLevels
        self.evalFreq: List[float] = evalFreq
        # constraints and masked windows
        self.constraints: Dict = {}
        self.insideOut: Dict = {}
        self.maskWindows: Dict = {}
        # stats is the stats to use
        self.stats: List[str] = []
        if "stats" in kwargs:
            self.stats = kwargs["stats"]
            self.prepareDicts()

    def setStats(self, stats):
        """Set the statistics to use for the constraints

        Parameters
        ----------
        stats : List[str]
            A list of statistics for which constraints will be provided
        """

        self.stats = stats
        self.prepareDicts()

    def addConstraint(self, stat: str, constraint, insideOut=[]) -> None:
        """Add a constraint for all levels and evaluation frequencies

        Parameters
        ----------
        stat : str
            The statistic to be constrained
        constraint : 
            The constraint parameter
        insideOut : (optional) 
            Inside out parameters
        """

        for ilevel in range(0, self.numLevels):
            self.addConstraintLevel(stat, constraint, ilevel, insideOut)

    def addConstraintLevel(
        self, stat: str, constraint, declevel: int, insideOut=[]
    ) -> None:
        """Add a constraint for a whole decimation level

        Parameters
        ----------
        stat : str
            The statistic to be constrained
        constraint : 
            The constraint parameter
        declevel : int
            The decimation level
        insideOut : (optional) 
            Inside out parameters
        """

        for eIdx, eFreq in enumerate(self.evalFreq[declevel]):
            self.addConstraintFreq(stat, constraint, declevel, eIdx, insideOut)

    def addConstraintFreq(
        self, stat: str, constraint, declevel: int, eIdx: int, insideOut=[]
    ) -> None:
        """Add a constraint for an evaluation frequency

        Parameters
        ----------
        stat : str
            The statistic to be constrained
        constraint : 
            The constraint parameter
        declevel : int
            The decimation level
        eIdx : int
            Evaluation Frequency index
        insideOut : (optional) 
            Inside out parameters
        """

        eFreq = self.evalFreq[declevel][eIdx]
        # insideOut = []
        # if "insideOut" in kwargs:
        #     insideOut = kwargs["insideOut"]
        for key in constraint:
            self.constraints[eFreq][stat][key] = constraint[key]
            if key in insideOut:
                self.insideOut[eFreq][stat][key] = True
            else:
                self.insideOut[eFreq][stat][key] = False

    def getConstraintFreq(self, declevel: int, eIdx: int, stat: str = ""):
        """Get constraints for an evaluation frequency

        Parameters
        ----------
        declevel : int
            The decimation level
        eIdx : int
            Evaluation Frequency index
        stat : str
            The statistic for which to return the constraints

        Returns
        -------
        out : list
            Constraints for the evaluation frequency
        """

        eFreq = self.evalFreq[declevel][eIdx]
        if stat != "":
            return self.constraints[eFreq][stat]
        return self.constraints[eFreq]

    def getMaskWindowsFreq(self, declevel: int, eIdx: int) -> Set[int]:
        """Get a set of the masked windows (used by WindowSelector)

        Parameters
        ----------
        declevel : int
            The decimation level
        eIdx : int
            Evaluation Frequency index

        Returns
        -------
        out : Set[int]
            A set with the indices of the masked windows
        """

        eFreq = self.evalFreq[declevel][eIdx]
        return self.maskWindows[eFreq]

    def prepareDicts(self) -> None:
        """Prepare dictionaries for storing masking information"""

        for ilevel in range(0, self.numLevels):
            for eFreq in self.evalFreq[ilevel]:
                # empty set for maskWindows to start off with
                # i.e. remove none
                self.maskWindows[eFreq] = set()
                self.constraints[eFreq] = {}
                self.insideOut[eFreq] = {}
                for stat in self.stats:
                    self.constraints[eFreq][stat] = {}
                    self.insideOut[eFreq][stat] = {}

    def resetMaskWindows(self) -> None:
        """Reset/Clear the masked windows"""

        for ilevel in range(0, self.numLevels):
            for eFreq in self.evalFreq[ilevel]:
                # empty set for maskWindows to start off with
                # i.e. remove none
                self.maskWindows[eFreq] = set()

    def view(self, declevel: int = 0, **kwargs) -> plt.figure:
        """Produces a 2-D plot with all the masked windows along the bottom for a decimation level
        
        Parameters
        ----------
        declevel : int, optional
            The decimation level. Default is 0.
        fig : matplotlib.pyplot.figure, optional
            A figure object
        plotfonts : Dict, optional
            A dictionary of plot fonts            
        """

        fig: plt.figure = (
            plt.figure(kwargs["fig"].number)
            if "fig" in kwargs
            else plt.figure(figsize=(10, 10))
        )
        plotfonts = kwargs["plotfonts"] if "plotfonts" in kwargs else getViewFonts()

        evalFreq = self.evalFreq[declevel]
        numFreq = len(evalFreq)
        freqIndex = range(0, numFreq)
        levelMasked: Set = set()
        numMasked: List[int] = list()
        freqLabel: List[str] = list()
        for eFreq in evalFreq:
            levelMasked.update(self.maskWindows[eFreq])
            numMasked.append(len(self.maskWindows[eFreq]))
            freqLabel.append("{:.5f}".format(eFreq))
        
        numMaskedTotal = len(levelMasked)
        levelMasked = np.array(sorted(list(levelMasked)))
        global2localMap = {}
        for idx in range(0, levelMasked.size):
            global2localMap[levelMasked[idx]] = idx
        # make array
        data = np.zeros(shape=(levelMasked.size, numFreq), dtype=int)
        for idx, eFreq in enumerate(evalFreq):
            for gI in self.maskWindows[eFreq]:
                lI = global2localMap[gI]
                data[lI, idx] = idx + 1

        st = fig.suptitle(
            "Masked windows for decimation level: {}".format(declevel),
            fontsize=plotfonts["suptitle"],
        )
        st.set_y(0.98)

        # now plot
        ax = plt.subplot(2, 1, 1)
        cmap = cm.get_cmap("Paired", numFreq)
        cmap.set_under("white")
        plt.pcolor(np.transpose(data), cmap=cmap, vmin=1, vmax=numFreq + 1)
        ax.set_yticks(np.array(freqIndex) + 0.5)
        ax.set_yticklabels(freqLabel)
        plt.ylabel(
            "Masked windows for decimation level", fontsize=plotfonts["axisLabel"]
        )
        plt.ylabel("Evaluation Frequency [Hz]", fontsize=plotfonts["axisLabel"])
        plt.title(
            "Relative masked windows for decimation level {}".format(declevel),
            fontsize=plotfonts["title"],
        )
        # set tick sizes
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(plotfonts["axisTicks"])
        # colourbar
        cb = plt.colorbar()
        cb.ax.set_title("Masked\nwindows")
        cb.set_ticks(np.arange(1, numFreq + 1) + 0.5)
        cb.set_ticklabels(freqLabel)

        # bar graph
        ax = plt.subplot(2, 1, 2)
        # add the total number masked
        freqIndex = range(0, numFreq + 1)
        numMasked.append(numMaskedTotal)
        freqLabel.append("Total masked for\n decimation level")
        plt.bar(freqIndex, numMasked)
        ax.set_xticks(freqIndex)
        ax.set_xticklabels(freqLabel)
        plt.xlabel("Evaluation Frequency [Hz]", fontsize=plotfonts["axisLabel"])
        plt.ylabel("Number of masked windows", fontsize=plotfonts["axisLabel"])
        plt.title(
            "Number of masked windows per evaluation frequency",
            fontsize=plotfonts["title"],
        )
        # set tick sizes
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(plotfonts["axisTicks"])

        fig.tight_layout(rect=[0.02, 0.02, 0.98, 0.92])
        plt.show()

        return fig

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst: List[str] = []
        textLst.append("Mask name = {}".format(self.maskName))
        textLst.append("Sampling frequency = {:.6f}".format(self.sampleFreq))
        textLst.append(
            "Statistics to use for constraints = {}".format(", ".join(self.stats))
        )
        textLst.append("Evaluation frequencies [Hz]")
        for il in range(0, self.numLevels):
            freqForLevel = self.evalFreq[il]
            for eFreq in freqForLevel:
                textLst.append(
                    "Decimation level = {:d}, Frequency = {:.6f} [Hz], number of masked windows = {:d}".format(
                        il, eFreq, len(self.maskWindows[eFreq])
                    )
                )
        return textLst

    def printConstraints(self) -> None:
        """Print information about the constraints"""

        blockPrint(
            "{} Constraints".format(self.__class__.__name__),
            self.printConstraintsList(),
        )

    def printConstraintsList(self) -> List[str]:
        """Constraint information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        evalFreq = sorted(list(self.constraints.keys()))
        for eFreq in evalFreq:
            textLst.append("Frequency = {:.6f} [Hz]".format(eFreq))
            for stat in self.stats:
                textLst.append("\tStatistic = {}".format(stat))
                # check to see if any constraints
                if len(list(self.constraints[eFreq][stat].keys())) == 0:
                    textLst.append("\tNo constraints for this statistic")
                    continue
                # if there are, print out
                for component in self.constraints[eFreq][stat]:
                    minVal = self.constraints[eFreq][stat][component][0]
                    maxVal = self.constraints[eFreq][stat][component][1]
                    textLst.append(
                        "\t{}\t{}\t{}\t{}".format(
                            component,
                            minVal,
                            maxVal,
                            self.insideOut[eFreq][stat][component],
                        )
                    )
        return textLst



from typing import List

# import from package
from resistics.utilities.utilsPrint import (
    generalPrint,
    warningPrint,
    errorPrint,
    blockPrint,
)


class Calculator(object):
    """Parent class for calculators 

    Parent class to ensure consistency of print methods

    Methods
    -------
    __repr__()
        Print status information
    __str__()
        Print status information
    printInfo()
        Print status information
    printList()
        Return a list of strings with useful information
    printText(infoStr)
        Print information to console
    printWarning(warnStr)
        Print a warning to the console
    printError(errorStr, quitRun=False)
        Print an error to the console and optionally quit execution       
    """

    def __repr__(self) -> str:
        """Print information"""

        return "\n".join(self.printList())

    def __str__(self) -> str:
        """Print information"""

        return self.__repr__()

    def printInfo(self) -> None:
        """Print information"""

        blockPrint(self.__class__.__name__, self.printList())

    def printList(self) -> List:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        return [""]

    def printText(self, infoStr: str) -> None:
        """General print to terminal

        Parameters
        ----------
        infoStr : str
            The string to print to the console
        """         

        generalPrint("{} Info".format(self.__class__.__name__), infoStr)

    def printWarning(self, warnStr: str) -> None:
        """Warning print to terminal
        
        Parameters
        ----------
        warnStr : str
            The string to print to the console
        """   

        warningPrint("{} Warning".format(self.__class__.__name__), warnStr)

    def printError(self, errorStr: str, quitRun: bool = False) -> None:
        """Error print to terminal and possibly quit

        Parameters
        ----------
        errorStr : str
            The string to print to the console
        quitRun : bool, optional (False)
            If True, the code will exit
        """          

        errorPrint("{} Error".format(self.__class__.__name__), errorStr, quitRun=quitRun)            


"""A package for the processing of magnetotelluric data

resistics is a package for the robust processing of magnetotelluric data. It includes several features focussed on traceability and data investigation. For more information, visit the package website at:

www.resistics.io
"""

__name__ = "resistics"
# short X.Y version
xyversion = "0.0.5"
# release
release = "rc1"
# combined version
__version__ = "{}{}".format(xyversion, release)



import sys
import os
import numpy as np
from datetime import datetime
from typing import Union

# import from package
from resistics.dataObjects.projectData import ProjectData
from resistics.dataObjects.statisticData import StatisticData
from resistics.dataObjects.maskData import MaskData
from resistics.calculators.windowSelector import WindowSelector
from resistics.calculators.maskCalculator import MaskCalculator
from resistics.ioHandlers.statisticIO import StatisticIO
from resistics.ioHandlers.maskIO import MaskIO
from resistics.utilities.utilsChecks import parseKeywords
from resistics.utilities.utilsProject import (
    projectText,
    getDecimationParameters,
    getWindowParameters,
)



def newMaskData(projData: ProjectData, sampleFreq: float) -> MaskData:
    """Get a mask data object that can then be passed onto calculateMask

    Parameters
    ----------
    projData : ProjectData
        A ProjectData instance
    sampleFreq : float
        The sampling frequency to mask
    
    Returns
    -------
    MaskData
        A mask data object with parameters set
    """

    decParams = getDecimationParameters(sampleFreq, projData.config)
    decParams.printInfo()
    return MaskData(
        decParams.sampleFreq, decParams.numLevels, decParams.evalFreqPerLevel
    )


def getMaskData(
    projData: ProjectData,
    site: str,
    maskName: str,
    sampleFreq: Union[float, int],
    **kwargs
) -> MaskData:
    """Get a mask data object

    Parameters
    ----------
    projData : projectData
        A project instance
    site : str
        The site for which to get the mask
    maskName : str
        The name of the mask
    sampleFreq : int, float
        The sampling frequency for which the mask was created
    specdir : str
        The spectra directory for which the mask was created  

    Returns
    -------
    MaskData
        A mask data object with the mask information
    """

    options = {}
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options = parseKeywords(options, kwargs)

    siteData = projData.getSiteData(site)
    maskIO = MaskIO(siteData.getSpecdirMaskPath(options["specdir"]))
    maskData = maskIO.read(maskName, sampleFreq)
    return maskData


def calculateMask(projData: ProjectData, maskData: MaskData, **kwargs):
    """Calculate masks sites

    Parameters
    ----------
    projData : projectData
        A project instance
    maskData : MaskData
        A mask data instance
    sites : List[str], optional
        A list of sites to calculate statistics for
    specdir : str, optional
        The spectra directory for which to calculate statistics
    """

    options = {}
    options["sites"] = projData.getSites()
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options = parseKeywords(options, kwargs)

    # create a maskCalculator object
    maskCalc = MaskCalculator(projData, maskData, specdir=options["specdir"])
    maskIO = MaskIO()
    sampleFreq = maskData.sampleFreq

    # loop over sites
    for site in options["sites"]:
        # see if there is a sample freq
        siteData = projData.getSiteData(site)
        siteSampleFreqs = siteData.getSampleFreqs()
        if sampleFreq not in siteSampleFreqs:
            continue

        # decimation and window parameters
        decParams = getDecimationParameters(sampleFreq, projData.config)
        decParams.printInfo()
        winParams = getWindowParameters(decParams, projData.config)

        # clear previous windows from maskCalc
        maskCalc.clearMaskWindows()
        # calculate masked windows
        maskCalc.applyConstraints(site)
        maskCalc.maskData.printInfo()

        # write maskIO file
        maskIO.datapath = os.path.join(siteData.getSpecdirMaskPath(options["specdir"]))
        maskIO.write(maskCalc.maskData)

        # test with the window selector
        winSelector = WindowSelector(
            projData, sampleFreq, decParams, winParams, specdir=options["specdir"]
        )
        winSelector.setSites([site])
        winSelector.addWindowMask(site, maskData.maskName)
        winSelector.calcSharedWindows()
        winSelector.printInfo()
        winSelector.printDatetimeConstraints()
        winSelector.printWindowMasks()
        winSelector.printSharedWindows()
        winSelector.printWindowsForFrequency()




import sys
import os
import numpy as np
from datetime import datetime
import matplotlib.pyplot as plt
from typing import Dict, Union, List

# import from package
from resistics.dataObjects.projectData import ProjectData
from resistics.calculators.calibrator import Calibrator
from resistics.ioHandlers.dataReader import DataReader
from resistics.ioHandlers.dataWriterInternal import DataWriterInternal
from resistics.utilities.utilsChecks import parseKeywords, isElectric
from resistics.utilities.utilsPlotter import savePlot, plotOptionsTime
from resistics.utilities.utilsProject import (
    projectText,
    projectBlock,
    checkDateOptions,
    getCalibrator,
    applyPolarisationReversalOptions,
    applyCalibrationOptions,
    applyFilterOptions,
    applyInterpolationOptions,
    applyNormaliseOptions,
    applyNotchOptions,
    applyResampleOptions,
)


def getTimeReader(projData: ProjectData, site: str, meas: str) -> DataReader:
    """Get a data reader object for a measurement

    Data readers will then return a timeData object when given a sample range of data to return for

    Parameters
    ----------
    projData : ProjectData
        A project data object
    site : str
        The site for which to get the data reader
    meas : str
        The measurement in the site for which to get the data reader

    Returns
    -------
    DataReader
        A data reader object which allows a user to get time data
    """

    siteData = projData.getSiteData(site)
    return siteData.getMeasurement(meas)


def preProcess(projData: ProjectData, **kwargs) -> None:
    """Pre-process project time data

    Preprocess the time data using filters, notch filters, resampling or interpolation. A new measurement folder is created under the site. The name of the new measurement folder is:
    prepend_[name of input measurement]_postpend. By default, prepend is "proc" and postpend is empty. 

    Processed time series data can be saved in a new site by using the outputsite option.

    Parameters
    ----------
    projData : ProjectData
        A project data object
    sites : str, List[str], optional
        Either a single site or a list of sites
    sampleFreqs : int, float, List[float], optional
        The frequencies to preprocess
    start : str, optional
        Start date of data to preprocess in format "%Y-%m-%d %H:%M:%S"
    stop : str, optional
        Stop date of data to process in format "%Y-%m-%d %H:%M:%S"
    outputsite : str, optional
        A site to output the preprocessed time data to. If this site does not exist, it will be created
    polreverse :  Dict[str, bool]
        Keys are channels and values are boolean flags for reversing        
    calibrate : bool, optional
        Boolean flag for calibrating the data. Default is false and setting to True will calibrate where files can be found.
    normalise : bool, optional
        Boolean flag for normalising the data. Default is false and setting to True will normalise each channel individually.
    filter : Dict, optional
        Filtering options in a dictionary
    notch : List[float], optional
        List of frequencies to notch in spectra given as a list of floats
    resamp : Dict, optional
        Resampling parameters in a dictionary with entries in the format: {sampleRateFrom: sampleRateTo}. All measurement directories of sampleRateFrom will be resampled to sampleRateTo
    interp : bool, optional
        Boolean flag for interpolating the data on to the second, so that sampling is coincident with seconds. This is not always the case. For example, SPAM data is not necessarily sampled on the second, whereas ATS data is. This function is useful when combining data of multiple formats. Interpolation does not change the sampling rate. Default is False.
    prepend : str, optional
        String to prepend to the output folder
    postpend : str, optional
        String to postpend to the output folder
    """

    options: Dict = {}
    options["sites"]: List = projData.getSites()
    options["sampleFreqs"]: List[float] = projData.getSampleFreqs()
    options["start"]: Union(bool, str) = False
    options["stop"]: Union(bool, str) = False
    options["outputsite"]: str = ""
    options["polreverse"]: Union[bool, Dict[str, bool]] = False    
    options["calibrate"]: bool = False
    options["normalise"]: bool = False
    options["filter"]: Dict = {}
    options["notch"]: List[float] = []
    options["resamp"]: Dict = {}
    options["interp"]: bool = False
    options["prepend"]: str = "proc"
    options["postpend"]: str = ""
    options = parseKeywords(options, kwargs)

    # print info
    text: List = ["Processing with options"]
    for op, val in options.items():
        text.append("\t{} = {}".format(op, val))
    projectBlock(text)

    # outputting to another site
    if options["outputsite"] != "":
        projectText(
            "Preprocessed data will be saved to output site {}".format(
                options["outputsite"]
            )
        )
        # create the site
        projData.createSite(options["outputsite"])
        projData.refresh()
        outputSitePath = projData.getSiteData(options["outputsite"]).timePath

    # output naming
    outPre = options["prepend"] + "_" if options["prepend"] != "" else ""
    outPost = "_" + options["postpend"] if options["postpend"] != "" else ""
    if outPre == "" and outPost == "" and options["outputsite"] == "":
        outPre = "proc_"

    # create a data calibrator writer instance
    cal = Calibrator(projData.calPath)
    if options["calibrate"]:
        cal.printInfo()
    writer = DataWriterInternal()

    # format dates
    if options["start"]:
        options["start"] = datetime.strptime(options["start"], "%Y-%m-%d %H:%M:%S")
    if options["stop"]:
        options["stop"] = datetime.strptime(options["stop"], "%Y-%m-%d %H:%M:%S")

    for site in options["sites"]:
        siteData = projData.getSiteData(site)
        siteData.printInfo()
        # loop over frequencies
        for sampleFreq in options["sampleFreqs"]:
            measurements = siteData.getMeasurements(sampleFreq)
            if len(measurements) == 0:
                # no data files at this sample rate
                continue

            # otherwise, process
            for meas in measurements:
                # get the reader
                projectText("Processing site {}, measurement {}".format(site, meas))
                reader = siteData.getMeasurement(meas)
                startTime = reader.getStartDatetime()
                stopTime = reader.getStopDatetime()
                if (options["start"] or options["stop"]) and not checkDateOptions(
                    options, startTime, stopTime
                ):
                    continue
                # if the data contributes, copy in the data if relevant
                if options["start"]:
                    startTime = datetime.strptime(options["start"], "%Y-%m-%d %H:%M:%S")
                if options["stop"]:
                    stopTime = datetime.strptime(options["stop"], "%Y-%m-%d %H:%M:%S")

                # calculate the samples
                sampleStart, sampleEnd = reader.time2sample(startTime, stopTime)
                # now get the data
                timeData = reader.getPhysicalSamples(
                    startSample=sampleStart, endSample=sampleEnd
                )
                timeData.printInfo()
                headers = reader.getHeaders()
                chanHeaders, chanMap = reader.getChanHeaders()

                # apply options
                applyPolarisationReversalOptions(options, timeData)
                applyCalibrationOptions(options, cal, timeData, reader)
                applyFilterOptions(options, timeData)
                applyNotchOptions(options, timeData)
                applyInterpolationOptions(options, timeData)
                applyResampleOptions(options, timeData)
                applyNormaliseOptions(options, timeData)

                # output dataset path
                if options["outputsite"] != "":
                    timePath = outputSitePath
                else:
                    timePath = siteData.timePath
                outPath = os.path.join(timePath, "{}{}{}".format(outPre, meas, outPost))
                # write time data - need to manually change some headers (hence the keywords)
                writer = DataWriterInternal()
                writer.setOutPath(outPath)
                writer.writeData(
                    headers,
                    chanHeaders,
                    timeData,
                    start_time=timeData.startTime.strftime("%H:%M:%S.%f"),
                    start_date=timeData.startTime.strftime("%Y-%m-%d"),
                    stop_time=timeData.stopTime.strftime("%H:%M:%S.%f"),
                    stop_date=timeData.stopTime.strftime("%Y-%m-%d"),
                    numSamples=timeData.numSamples,
                    sample_freq=timeData.sampleFreq,
                    physical=True,
                )
                writer.printInfo()


def viewTime(
    projData: ProjectData, startDate: str, endDate: str, **kwargs
) -> Union[plt.figure, None]:
    """View timeseries in the project

    Parameters
    ----------
    startDate : str
        The start of the data range to plot
    endDate : str
        The end of the date range to plot
    sites : List[str], optional
        List of sites 
    sampleFreqs : List[float], optional
        List of sample frequencies to plot
    chans : List[str], optional
        List of channels to plot
    polreverse :  Dict[str, bool]
        Keys are channels and values are boolean flags for reversing        
    calibrate : bool, optional
        Boolean flag to calibrate data
    normalise : bool, optional
        Boolean flag to normalise the data
    notch : List[float], optional
        List of frequencies to notch out
    filter : Dict, optional
        Filter parameters
    show : bool, optional
        Boolean flag to show the plot
    save : bool, optional
        Boolean flag to save the plot to images folder
    plotoptions : Dict
        Dictionary of plot options

    Returns
    -------
    matplotlib.pyplot.figure or None
        A matplotlib figure unless the plot is not shown and is saved, in which case None and the figure is closed.
    """

    # default options
    options = {}
    options["sites"]: List[str] = projData.sites
    options["sampleFreqs"]: Union[List[float], List[str]] = projData.getSampleFreqs()
    options["chans"]: List[str] = ["Ex", "Ey", "Hx", "Hy", "Hz"]
    options["polreverse"]: Union[bool, Dict[str, bool]] = False    
    options["calibrate"]: bool = False
    options["normalise"]: bool = False
    options["filter"]: Dict = {}
    options["notch"]: List[float] = []
    options["show"]: bool = True
    options["save"]: bool = False
    options["plotoptions"]: Dict = plotOptionsTime()
    options = parseKeywords(options, kwargs)

    # prepare calibrator
    cal = getCalibrator(projData.calPath, projData.config)
    if options["calibrate"]:
        cal.printInfo()

    # format startDate and endDate
    start = datetime.strptime("{}.000".format(startDate), "%Y-%m-%d %H:%M:%S.%f")
    stop = datetime.strptime("{}.000".format(endDate), "%Y-%m-%d %H:%M:%S.%f")
    # collect relevant data - dictionary to store timeData
    timeDataAll = {}
    for site in options["sites"]:
        siteData = projData.getSiteData(site)
        siteData.printInfo()
        measurements = siteData.getMeasurements()
        timeDataAll[site] = {}

        # loop over measurements and save data for each one
        for meas in measurements:
            sampleFreq = siteData.getMeasurementSampleFreq(meas)
            if sampleFreq not in options["sampleFreqs"]:
                continue

            # check if data contributes to user defined time period
            siteStart = siteData.getMeasurementStart(meas)
            siteStop = siteData.getMeasurementEnd(meas)
            if siteStop < start or siteStart > stop:
                continue

            reader = siteData.getMeasurement(meas)
            # get the samples of the datetimes
            sampleStart, sampleStop = reader.time2sample(start, stop)
            # as the samples returned from time2sample are rounded use sample2time to get the appropriate start and end times for those samples
            readStart, readStop = reader.sample2time(sampleStart, sampleStop)
            # get the data
            timeData = reader.getPhysicalData(
                readStart, readStop, chans=options["chans"]
            )

            projectText(
                "Plotting measurement {} of site {} between {} and {}".format(
                    meas, site, readStart, readStop
                )
            )

            # apply various options
            applyPolarisationReversalOptions(options, timeData)
            applyCalibrationOptions(options, cal, timeData, reader)
            applyFilterOptions(options, timeData)
            applyNotchOptions(options, timeData)
            applyNormaliseOptions(options, timeData)
            timeDataAll[site][meas] = timeData

    # plot all the data
    plotfonts = options["plotoptions"]["plotfonts"]
    fig = plt.figure(figsize=options["plotoptions"]["figsize"])
    for site in timeDataAll:
        for meas in timeDataAll[site]:
            timeData = timeDataAll[site][meas]
            timeData.view(
                sampleStop=timeDataAll[site][meas].numSamples - 1,
                fig=fig,
                chans=options["chans"],
                label="{} - {}".format(site, meas),
                xlim=[start, stop],
                plotfonts=plotfonts,
            )

    # add the suptitle
    st = fig.suptitle(
        "Time data from {} to {}".format(
            start.strftime("%Y-%m-%d %H-%M-%S"), stop.strftime("%Y-%m-%d %H-%M-%S")
        ),
        fontsize=plotfonts["suptitle"],
    )
    st.set_y(0.98)

    # do the axis labels
    numChans = len(options["chans"])
    for idx, chan in enumerate(options["chans"]):
        plt.subplot(numChans, 1, idx + 1)
        # do the yaxis
        if isElectric(chan):
            plt.ylabel("mV/km", fontsize=plotfonts["axisLabel"])
            if len(options["plotoptions"]["Eylim"]) > 0:
                plt.ylim(options["plotoptions"]["Eylim"])
        else:
            if options["calibrate"]:
                plt.ylabel("nT", fontsize=plotfonts["axisLabel"])
            else:
                plt.ylabel("mV", fontsize=plotfonts["axisLabel"])
            if len(options["plotoptions"]["Hylim"]) > 0:
                plt.ylim(options["plotoptions"]["Hylim"])
        plt.legend(loc=1, fontsize=plotfonts["legend"])

    # plot format
    fig.tight_layout(rect=[0, 0.02, 1, 0.96])
    fig.subplots_adjust(top=0.92)

    # plot show and save
    if options["save"]:
        impath = projData.imagePath
        filename = "timeData_{}_{}".format(
            start.strftime("%Y-%m-%d_%H-%M-%S_"), stop.strftime("%Y-%m-%d_%H-%M-%S")
        )
        savename = savePlot(impath, filename, fig)
        projectText("Image saved to file {}".format(savename))
    if options["show"]:
        plt.show(block=options["plotoptions"]["block"])
    if not options["show"] and options["save"]:
        plt.close(fig)
        return None
    return fig



import numpy as np
import math
from typing import List


def getEvaluationFreq(fs: float, minFreq: float) -> np.ndarray:
    """Calculate evaluation frequencies with mimum allowable frequency

    Highest frequency is Nyquist/2. Lowest allowable frequency is provided by user
    f_i = f_max / pow(2,(i-1)/2)
    
    Parameters
    ----------
    fs : float
        Sampling frequency
    minFreq : float
        Minimum allowable frequency    

    Returns
    -------
    out : np.ndarray
        Array of evaluation frequencies
    """

    fmax: float = fs / 4
    freq: List = []
    i: int = 1
    f: float = fmax
    while f > minFreq:
        f = fmax / math.pow(2, (i - 1.0) / 2.0)
        freq.append(f)
        i = i + 1
    return np.array(freq)


def getEvaluationFreqSize(fs: float, numFreq: int) -> np.ndarray:
    """Calculate evaluation frequencies with maximum size

    Highest frequency is Nyquist/2 and number of evaluation frequencies is set by user
    f_i = f_max / pow(2,(i-1)/2)
    
    Parameters
    ----------
    fs : float
        Sampling frequency
    numFreq : int
        Number of evaluation frequencies    

    Returns
    -------
    out : np.ndarray
        Array of evaluation frequencies
    """

    fmax: float = fs / 4
    freq: List = []
    i: int = 1
    f: float = fmax
    while i <= numFreq:
        f = fmax / math.pow(2, (i - 1.0) / 2.0)
        freq.append(f)
        i = i + 1
    return np.array(freq)




import os
from datetime import datetime, timedelta
import matplotlib.pyplot as plt
from typing import List, Dict, Union, ClassVar

# import from package
from resistics.dataObjects.dataObject import DataObject
from resistics.dataObjects.siteData import SiteData
from resistics.dataObjects.configData import ConfigData
from resistics.utilities.utilsIO import (
    checkAndMakeDir,
    checkDirExistence,
    getDirsInDirectory,
    getDataDirsInDirectory,
    getFilesInDirectory,
)
from resistics.utilities.utilsPrint import arrayToString, listToString
from resistics.utilities.utilsReader import getDataReader


class ProjectData(DataObject):
    """Project class for storing project attributes

    Attributes
    ----------
    projectFile : str
        Path to project file
    timePath : str
        Path to time series data
    specPath : str
        Path to spectra data
    statPath : str
        Path to statistics
    transPath : str
        Path to transfer function data
    calPath : str
        Path to calibration data
    imagePath : str
        Path to project images
    refTime : datetime.datetime
        The project reference time
    projStart : datetime.datetime
        The first time sample in the data
    projEnd : datetime.datetime
        The last time sample in the data
    sites : List[str]
        List of sites in the project
    siteData : Dict[str, SiteData]
        Site data for each site        
    calFiles : List
        A list of calibration files
    config : ConfigData
        Configuration data

    Methods
    -------
    __init__()
        Initialise project with default values
    refresh()
        Refresh the project
    checkDirectories()
        Check the project directories to make sure they exist
    getCalInfo()
        Get calibration file information
    getSiteInfo()   
        Get information about all the sites and store site data information
    getSites(fs=None)
        Get a list of all sites, or those with data at a sampling frequency of fs Hz 
    getNumSites()
        Get the total number of sites
    getSampleFreqs()
        Get a list of all the unique sample frequencies in the project
    getSiteConcurrent(site)
        Get a list of sites that were recording concurrently to site
    getSiteData(site)
        Get site data for a single site
    checkSite(site)
        Check a site exists
    view()
        View the project timeline
    printList()
        Class status returned as list of strings   
    """

    def __init__(
        self,
        projectFile: str,
        refTime: datetime,
        calPath: str,
        timePath: str,
        specPath: str,
        statPath: str,
        maskPath: str,
        transFuncPath: str,
        imagePath: str,
        config=None,
    ):
        # project file and reference time
        self.projectFile: str = projectFile
        self.refTime: datetime = refTime
        # paths to various parts of the project
        self.calPath: str = calPath
        self.timePath: str = timePath
        self.specPath: str = specPath
        self.statPath: str = statPath
        self.maskPath: str = maskPath
        self.transFuncPath: str = transFuncPath
        self.imagePath: str = imagePath
        self.projStart: datetime = datetime.now()
        self.projEnd: datetime = datetime.now()
        # site information
        self.sites: List[str] = []
        self.siteData: Dict[str, SiteData] = {}
        # cal files list
        self.calFiles: List = []
        # configuration options
        self.config = config
        if self.config is None:
            self.config = ConfigData()
        # refresh
        self.refresh()

    def refresh(self) -> None:
        """Refresh the project

        If new time, spectra or other data directories have been added to the project, these will not show up in the project information until the project has been refreshed (i.e. all the information read again from the disc)
        """

        self.checkDirectories()
        self.getCalInfo()
        self.getSiteInfo()
        # update start and end dates for the project
        startDates = []
        endDates = []
        for site in self.sites:
            startDates.append(self.getSiteData(site).siteStart)
            endDates.append(self.getSiteData(site).siteEnd)
        if len(self.sites) > 0:
            self.projStart = min(startDates)
            self.projEnd = max(endDates)

    def checkDirectories(self) -> None:
        """Check to see if project directories exist and if not, make them"""

        # data directories
        checkAndMakeDir(self.timePath)
        checkAndMakeDir(self.specPath)
        checkAndMakeDir(self.statPath)
        checkAndMakeDir(self.maskPath)
        checkAndMakeDir(self.transFuncPath)
        # calibration directories
        checkAndMakeDir(self.calPath)
        # image directory
        checkAndMakeDir(self.imagePath)

    def getCalInfo(self):
        """Get a list of all the calibration files in the project"""

        self.calFiles = getFilesInDirectory(self.calPath)

    def getSiteInfo(self) -> None:
        """Get information about sites"""

        self.sites = getDirsInDirectory(self.timePath)
        self.sites = sorted(self.sites)
        for site in self.sites:
            self.siteData[site] = SiteData(
                site,
                self.timePath,
                self.specPath,
                self.statPath,
                self.maskPath,
                self.transFuncPath,
            )

    def getSites(self, sampleFreq: Union[None, float, int] = None):
        """Get list of sites

        Optionally filtered to those with measurements recorded at fs Hz

        Parameters
        ----------
        sampleFreq : float, int, optional
            Sampling frequency Hz

        Returns
        -------
        List[str]
            List of site names
        """

        if sampleFreq is None:
            return self.sites
        # fs is supplied
        filteredSites = []
        for site in self.sites:
            if sampleFreq in self.getSiteData(site).getSampleFreqs():
                filteredSites.append(site)
        return filteredSites

    def getNumSites(self) -> int:
        """Number of sites in the project

        Returns
        -------
        int
            Number of sites
        """

        return len(self.sites)

    def getSampleFreqs(self) -> List[float]:
        """List of all the sampling frequencies found in the project

        Returns
        -------
        List[float]
            List of sampling frequencies in Hz
        """

        sampleFreq = set()
        for site in self.sites:
            sampleFreq.update(self.getSiteData(site).getSampleFreqs())
        return sorted(list(sampleFreq))

    def getSiteConcurrent(self, site: str) -> List[str]:
        """Find sites which were occupied at the same time as a chosen site

        This lists all sites that were occupied at the same time as the user supplied site. It can be useful in cases where multi-site processing will be performed

        Parameters
        ----------
        site : str
            The name of the site

        Returns
        -------
        List[str]
            List of sites which were recording concurrently
        """

        siteStart = self.getSiteData(site).siteStart
        siteEnd = self.getSiteData(site).siteEnd
        concSites = []
        for s in self.sites:
            if s == site:
                continue
            start = self.getSiteData(s).siteStart
            end = self.getSiteData(s).siteEnd
            # check that start time is before end time of site and that end time is after start time of site
            if (start < siteEnd) and (end > siteStart):
                concSites.append(s)
        return concSites

    def getSiteData(self, site) -> Union[SiteData, bool]:
        """Get site data for site

        Parameters
        ----------
        site : str
            Site name

        Returns
        -------
        SiteData
            Site data for site or False if the site is not found
        """

        if site not in self.sites:
            self.printWarning("Site {} does not exist in the project".format(site))
            return False
        return self.siteData[site]

    def checkSite(self, site: str) -> bool:
        """Check if site is part of the project
        
        Parameters
        ----------
        site : str
            Name of the site
        """

        if site in self.sites:
            return True
        else:
            return False

    def createSite(self, site: str) -> bool:
        """Creates a site folder in the time data path

        Parameters
        ----------
        site : str
            Name of the site
        """

        sitePath = os.path.join(self.timePath, site)
        chk = checkDirExistence(sitePath)
        if chk:
            self.printWarning(
                "Site {} already exists in project time data folder".format(site)
            )
            return False
        checkAndMakeDir(sitePath)
        return True

    def view(self, **kwargs) -> plt.figure:
        """Plot a timeline of the project
        
        Parameters
        ----------
        figsize : Tuple (width, height), optional
            The figure size
        plotfonts : Dict, optional
            Fonts to use for plotting
        show : bool, optional
            Boolean flag fow showing plot. Default is True.
        """

        import matplotlib.dates as mdates
        from matplotlib.patches import Rectangle
        import numpy as np
        from resistics.utilities.utilsPlotter import getPlotFonts

        figsize = kwargs["figsize"] if "figsize" in kwargs else (15, 8)
        plotFonts = kwargs["plotFonts"] if "plotFonts" in kwargs else getPlotFonts()
        show = kwargs["show"] if "show" in kwargs else True

        fig = plt.figure(figsize=figsize)
        ax = plt.subplot(1, 1, 1)
        # get xlimits
        xStart = self.refTime - timedelta(days=2)
        xEnd = self.projEnd + timedelta(days=2)

        numSites = len(self.sites)
        height = 0.4
        for idx, site in enumerate(self.sites):
            idx = numSites - idx
            siteStart = mdates.date2num(self.siteData[site].siteStart)
            siteEnd = mdates.date2num(self.siteData[site].siteEnd)
            siteWidth = siteEnd - siteStart
            yOffset = 0.5 + height
            rect = Rectangle((siteStart, idx - yOffset), siteWidth, 2 * height)
            ax.add_patch(rect)
        # plot the reference line
        plt.plot(
            [self.refTime, self.refTime], [0, numSites], color="lightcoral", ls="-"
        )
        plt.text(
            self.refTime, 0.02, "  Reference Time", fontsize=plotFonts["axisTicks"]
        )
        # x axis formatting
        weeks = mdates.WeekdayLocator()
        weeksFmt = mdates.DateFormatter("%Y-%m-%d")
        days = mdates.DayLocator()
        ax.xaxis.set_major_locator(weeks)
        ax.xaxis.set_major_formatter(weeksFmt)
        ax.xaxis.set_minor_locator(days)
        plt.xlim([xStart, xEnd])
        ax.xaxis.grid(True, which="minor", ls="--", color="gainsboro")
        plt.xlabel("Date", fontsize=plotFonts["axisLabel"])        
        fig.autofmt_xdate(rotation=30)
        for label in ax.get_xticklabels():
            label.set_fontsize(plotFonts["axisTicks"])        
        # y axis formatting
        plt.ylim([0, numSites])
        yticks = np.arange(1, numSites + 1)
        ax.set_yticks(yticks)
        ax.set_yticklabels([])
        ax.set_yticks(yticks - 0.5, minor=True)
        ax.set_yticklabels(
            self.sites[::-1], minor=True, fontsize=plotFonts["axisTicks"]
        )
        ax.yaxis.grid(True, which="major", ls="--", color="gainsboro")
        plt.ylabel("Sites", fontsize=plotFonts["axisLabel"], labelpad=20)
        # title
        plt.title("Project Timeline", fontsize=plotFonts["title"])

        fig.tight_layout()
        if show:
            plt.show()
        return fig

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst: List[str] = []
        textLst.append("Time data path = {}".format(self.timePath))
        textLst.append("Spectra data path = {}".format(self.specPath))
        textLst.append("Statistics data path = {}".format(self.statPath))
        textLst.append("Mask data path = {}".format(self.maskPath))
        textLst.append("TransFunc data path = {}".format(self.transFuncPath))
        textLst.append("Calibration data path = {}".format(self.calPath))
        textLst.append("Images data path = {}".format(self.imagePath))
        textLst.append(
            "Reference time = {}".format(self.refTime.strftime("%Y-%m-%d %H:%M:%S"))
        )
        textLst.append(
            "Project start time = {}".format(
                self.projStart.strftime("%Y-%m-%d %H:%M:%S.%f")
            )
        )
        textLst.append(
            "Project stop time = {}".format(
                self.projEnd.strftime("%Y-%m-%d %H:%M:%S.%f")
            )
        )
        textLst.append("Project found {} sites:".format(self.getNumSites()))
        for site in self.sites:
            textLst.append(
                "{}\t\tstart: {}\tend: {}".format(
                    site,
                    self.getSiteData(site).siteStart,
                    self.getSiteData(site).siteEnd,
                )
            )
        textLst.append(
            "Sampling frequencies found in project (Hz): {}".format(
                listToString(self.getSampleFreqs())
            )
        )
        return textLst




import os
import numpy as np
from datetime import datetime
from typing import List

# import from package
from resistics.ioHandlers.ioHandler import IOHandler
from resistics.dataObjects.spectrumData import SpectrumData
from resistics.utilities.utilsIO import checkAndMakeDir
from resistics.utilities.utilsPrint import breakComment, arrayToStringSci
from resistics.utilities.utilsMath import intdiv


class SpectrumWriter(IOHandler):
    """Writes spectra data for measurements

    Spectrum writer writes out spectra data as either binary (recommended) or ascii (more space required). A spectra file should be written out for each decimation level along with an information file, again for each decimation level. A single comments file is written out to record the processing sequence.

    Spectra files are located in:
    project -> specData -> site -> datapath -> spectra data files

    .dat files are ascii formatted data
    .bin files are binary formatted data

	Attributes
	----------
    datapath : str 
        Data root for spectra data
    filepath : str
        Filepath for spectra files
    refTime : datetime
        The reference time for the project
    file : bool 
        The data file 

	Methods
	-------
	__init__(datapath, refTime)
		Initialise the SpectrumWriter
    openBinaryForWriting(filename, fileInc, sampleFreq, winSize, winOverlap, globalOffset, numWindows, channels)
        Open file for writing binary spectra data
    writeBinary(specData)  
        Write out binary spectra data for a single time window          
    openAsciiForWriting(filename, fileInc, sampleFreq, winSize, winOverlap, globalOffset, numWindows, channels)
        Open file for writing ascii spectra data 
    writeAscii(specData)  
        Write out ascii spectra data for a single time window          
    writeInfoFile(filepath, sampleFreq, winSize, winOverlap, globalOffset, numWindows, channels)
        Write out the spectra information file
    writeCommentsFile(comments)   
        Write out the comments file      
    getFileBase(filename, fileInc)
        Get the spectra file name      
    closeFile()
        Close spectra data file 
    printList()
        Class status returned as list of strings       
	"""    

    def __init__(self, datapath: str, refTime: datetime):
        """Initialise spectrum writer 

        Parameters
        ----------
        datapath : str
            Root pathway for spectra data
        refTime : datetime
            Reference time
        """        

        self.datapath: str = datapath
        self.filepath: str = ""
        self.refTime: datetime = refTime
        self.file = None

    # def getDataRoot(self):
    #     return self.datapath

    # def getFilePath(self):
    #     return self.filepath

    # def getFile(self):
    #     return self.file

    # def getRefTime(self):
    #     return self.refTime

    # def setDataRoot(self, datapath):
    #     self.datapath = datapath

    def openBinaryForWriting(
        self,
        filename,
        fileInc,
        sampleFreq,
        winSize,
        winOverlap,
        globalOffset,
        numWindows,
        channels,
    ) -> None:
        """Write binary spectrum file 

        Parameters
        ----------
        filename : str
            Filename for spectra files
        fileInc : int
            The decimation level
        sampleFreq : float
            Sampling frequeny of time data
        winSize : int
            Window size in samples for time data windows
        winOverlap : int
            Overlap size in samples for time data windows
        globalOffset : int
            Global offset for local and global indices
        numWindows : int
            The number of windows in the time data
        channels : List[str]
            Channels in data
        """

        # sort channels alphabetically - matching the order in the data files
        self.channels = sorted(channels)

        checkAndMakeDir(self.datapath)
        filebase: str = filename + "{:02d}".format(fileInc)
        # info file
        filepathInfo: str = os.path.join(self.datapath, filebase + ".info")
        self.writeInfoFile(
            filepathInfo,
            sampleFreq,
            winSize,
            winOverlap,
            globalOffset,
            numWindows,
            channels,
        )
        # open file for data
        self.filepath: str = os.path.join(self.datapath, filebase + ".bin")
        self.printText("Opening file {}".format(self.filepath))
        self.file = open(self.filepath, "wb")

    def writeBinary(self, specData):
        """Write spectrum data to binary file 

        Parameters
        ----------
        specData : SpectrumData
            Spectrum data
        """

        for c in self.channels:
            # save as complex64 instead of 128 - otherwise too big
            self.file.write(specData.data[c].astype("complex64").tobytes())

    def openAsciiForWriting(
        self,
        filename: str,
        fileInc: str,
        sampleFreq: float,
        winSize: int,
        winOverlap: int,
        globalOffset: int,
        numWindows: int,
        channels: List[str],
    ) -> None:
        """Write ascii spectrum file 

        Parameters
        ----------
        filename : str
            Filename for spectra files
        fileInc : int
            The decimation level
        sampleFreq : float
            Sampling frequeny of time data
        winSize : int
            Window size in samples for time data windows
        winOverlap : int
            Overlap size in samples for time data windows
        globalOffset : int
            Global offset for local and global indices
        numWindows : int
            The number of windows in the time data
        channels : List[str]
            Channels in data
        """

        # sort channels alphabetically - matching the order in the data files
        self.channels = sorted(channels)

        checkAndMakeDir(self.datapath)
        filebase: str = filename + "{:02d}".format(fileInc)
        # info file
        filepathInfo: str = os.path.join(self.datapath, filebase + ".info")
        self.writeInfoFile(
            filepathInfo,
            sampleFreq,
            winSize,
            winOverlap,
            globalOffset,
            numWindows,
            channels,
        )
        # open file for data
        self.filepath: str = os.path.join(self.datapath, filebase + ".dat")
        self.printText("Opening file {}".format(self.filepath))
        self.file = open(self.filepath, "w")

    def writeAscii(self, specData: SpectrumData) -> None:
        """Write spectrum data to ascii file 

        Parameters
        ----------
        specData : SpectrumData
            Spectrum data
        """

        for c in self.channels:
            outStr = arrayToStringSci(specData.data[c])
            outStr = outStr + "\n"
            self.file.write(outStr)

    def writeInfoFile(
        self,
        filepath: str,
        sampleFreq: float,
        winSize: int,
        winOverlap: int,
        globalOffset: int,
        numWindows: int,
        channels: List[str],
    ) -> None:
        """Write info file 

        Parameters
        ----------
        filepath : str
            Filepath for info file
        sampleFreq : float
            Sampling frequeny of time data
        winSize : int
            Window size in samples for time data windows
        winOverlap : int
            Overlap size in samples for time data windows
        globalOffset : int
            Global offset for local and global indices
        numWindows : int
            The number of windows in the time data
        channels : List[str]
            Channels in data
        """

        infoFile = open(filepath, "w")
        # write out header information
        numChannels = len(channels)
        tmp = winSize + 1  # if winSize is odd, this will go down
        if winSize % 2 == 0:
            tmp = tmp + 1
        dataSize = intdiv(tmp, 2)
        infoFile.write(
            "Reference time = {}\nSample frequency = {:.8f}\nWindow size = {:d}\nWindow overlap = {:d}\nGlobal offset = {:d}\n".format(
                self.refTime.strftime("%Y-%m-%d %H:%M:%S.%f"),
                sampleFreq,
                winSize,
                winOverlap,
                globalOffset,
            )
        )
        infoFile.write(
            "Number of windows = {:d}\nData size = {:d}\n".format(numWindows, dataSize)
        )
        infoFile.write("Number of channels = {:d}\n".format(numChannels))
        infoFile.write("Channels = " + " ".join(channels))
        infoFile.close()

    def writeCommentsFile(self, comments: List[str]) -> None:
        """Write comments file 

        Parameters
        ----------
        comments : List[str]
            List of comments
        """

        filepathComments: str = os.path.join(self.datapath, "comments.txt")
        commentFile = open(filepathComments, "w")
        for comment in comments:
            commentFile.write("{}\n".format(comment))
        # add a comment about writing out
        commentFile.write(
            "Spectra data written out to {} on {}\n".format(
                self.datapath, datetime.now()
            )
        )
        commentFile.write(breakComment())
        commentFile.close()

    def getFileBase(self, filename: str, fileInc: int) -> str:
        """Read comments file from filepath

        Parameters
        ----------
        filename: str
            Filename of spectra files
        fileInc : int
            The decimation level
        """

        return filename + "{:02d}".format(fileInc)

    def closeFile(self):
        """Close spectra file"""

        if self.filepath != "" and self.file:
            self.printText("Closing file {}".format(self.filepath))
            self.file.close()
            self.filepath = ""
        else:
            print("No file open")

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Data root = {}".format(self.datapath))
        if not (self.file is None):
            textLst.append("Current file open: {}".format(self.filepath))
        return textLst




import os
import random
import numpy as np
import scipy.interpolate as interp
from typing import List, Dict, Tuple

# import from package
from resistics.calculators.calculator import Calculator
from resistics.ioHandlers.transferFunctionWriter import TransferFunctionWriter
from resistics.dataObjects.transferFunctionData import TransferFunctionData
from resistics.utilities.utilsIO import checkAndMakeDir, fileFormatSampleFreq
from resistics.utilities.utilsPrint import generalPrint, warningPrint, blockPrint
from resistics.utilities.utilsSmooth import smooth1d
from resistics.utilities.utilsRobust import (
    sampleMAD0,
    hermitianTranspose,
    olsModel,
    chatterjeeMachler,
    mmestimateModel,
)


class ProcessorSingleSite(Calculator):
    """Performs single site transfer function calculations 

    By default, the ProcessorSingleSite is setup to calculate the impedance tensor using Hx, Hy as input channels and Ex, Ey as output channels. To calculate the Tipper, the appropriate input and output channels have to be set.

    Attributes
    ----------
    winSelector : WindowSelect
        A window selector object which defines which windows to use in the linear model
    decParams : DecimationParameters
        DecimationParameters object with information about the decimation scheme
    winParams : WindowParameters
        WindowParameters object with information about the windowing
    outpath : str 
        Location to put the calculated transfer functions (Edi files)
    inSite : str 
        The site to use for the input channels 
    inChannels: List[str] (["Hx", "Hy"])
        List of hannels to use as input channels for the linear system
    inSize : int 
        Number of input channels
    outSite : str 
        The site to use for the output channels
    outChannels : List[str] (["Ex", "Ey"])
        List of channels to use as output channels for the linear system
    outSize : int
        Number of output channels
    allChannels : List[str] 
        inChannels and outChannels combined into a single list
    crossChannels : List[str] 
        The channels to calculate the cross spectra out for
    intercept : bool (default False)
        Flag for including an intercept (static) term in the linear system
    method : str (options, "ols", "cm") 
        String for describing what solution method to use
    win : str (default hanning)
        Window function to use in robust solution
    winSmooth : int (default -1)
        The size of the window smoother. If -1, this will be autocalculated based on data size
    postpend : str (default "")
        String to postpend to the output filename to help file management
    evalFreq : List[float] or np.ndarray
        The evaluation frequencies
    impedances : List

    variances : List

    Methods
    -------
    __init__(proj, maskData)
        Initialise with a Project instance and MaskData instance


    printList()
        Class status returned as list of strings
    """

    def __init__(self, winSelector, outpath: str):
        # required
        self.winSelector = winSelector
        self.decParams = winSelector.decParams
        self.winParams = winSelector.winParams
        self.outpath: str = outpath

        # default parameters for user options
        self.inSite: str = "dummy"
        self.inChannels: List[str] = ["Hx", "Hy"]
        self.inSize: int = len(self.inChannels)
        self.outSite: str = "dummy"
        self.outChannels: List[str] = ["Ex", "Ey"]
        self.outSize: int = len(self.outChannels)
        self.allChannels: List[str] = self.inChannels + self.outChannels
        self.crossChannels: List[str] = self.allChannels
        # solution options
        self.intercept: bool = False
        self.method: str = "cm"
        # smoothing options
        self.win: str = "hanning"
        self.winSmooth: int = -1
        # output filename
        self.postpend: str = ""
        # evaluation frequency data
        self.evalFreq = []
        self.impedances = []
        self.variances = []

    def setInput(self, inSite: str, inChannels: List[str]) -> None:
        """Set information about input site and channels
    
        Parameters
        ----------
        inSite : str
            Site to use for input channel data
        inChannels : List[str]
            Channels to use as the input in the linear system
        """

        self.inSite = inSite
        self.inChannels = inChannels
        self.inSize = len(inChannels)
        # set all and cross channels
        self.allChannels = self.inChannels + self.outChannels
        self.crossChannels = self.allChannels

    def setOutput(self, outSite: str, outChannels: List[str]) -> None:
        """Set information about output site and channels
    
        Parameters
        ----------
        inSite : str
            Site to use for output channel data
        inChannels : List[str]
            Channels to use as the output in the linear system
        """

        self.outSite = outSite
        self.outChannels = outChannels
        self.outSize = len(outChannels)
        # set all and cross channels
        self.allChannels = self.inChannels + self.outChannels
        self.crossChannels = self.allChannels

    def process(self) -> None:
        """Process spectra data

        The processing sequence for each decimation level is as below:

        1. Get shared (unmasked) windows for all relevant sites (inSite and outSite)
        2. For shared unmasked windows
            
            - Calculate out the spectral power data
            - Interpolate calculated spectral power data to the evaluation frequencies for the decimation level 
        
        3. For each evaluation frequency
            
            - Do the robust processing to calculate the transfer function at that evaluation frequency


        NOTES
        -----
        The spectral power data is smoothed as this tends to improve results. The smoothing can be changed by setting the smoothing parameters. This method is still subject to change in the future as it is an area of active work
        """

        numLevels: int = self.decParams.numLevels
        for iDec in range(0, numLevels):
            self.printText("Processing decimation level {}".format(iDec))
            fs = self.decParams.getSampleFreqLevel(iDec)
            # get the number of all shared windows and the number of unmasked windows
            # unmasked windows are ones that will actually be used in the calculation
            numWindows = self.winSelector.getNumSharedWindows(iDec)
            unmaskedWindows = self.winSelector.getUnmaskedWindowsLevel(iDec)
            numUnmasked = len(unmaskedWindows)
            self.printText(
                "Total shared windows for decimation level = {}".format(numWindows)
            )
            self.printText(
                "Total unmasked windows for decimation level = {}".format(numUnmasked)
            )
            if numUnmasked == 0:
                self.printText(
                    "No unmasked windows found at this decimation level ({:d}), continuing to next level".format(
                        iDec
                    )
                )
                continue  # continue to next decimation level
            self.printText("{} windows will be processed".format(numUnmasked))

            # get the evaluation frequencies
            evalFreq = self.decParams.getEvalFrequenciesForLevel(iDec)
            # set some variables
            totalSize: int = self.inSize + self.outSize
            numEvalFreq: int = len(evalFreq)
            dataSize: int = self.winSelector.getDataSize(iDec)
            freq: np.ndarray = np.linspace(0, fs / 2, dataSize)
            # get the window smoothing params
            smoothLen: int = self.getWindowSmooth(datasize=dataSize)

            # create the data array for each evaluation frequency
            # keep the spectral power information for all windows
            evalFreqData: np.ndarray = np.empty(
                shape=(numEvalFreq, numWindows, totalSize, totalSize), dtype="complex"
            )

            # an array for the window data
            winSpectraMatrix: np.ndarray = np.empty(
                shape=(totalSize, totalSize, dataSize), dtype="complex"
            )
            winDataArray: np.ndarray = np.empty(
                shape=(totalSize, dataSize), dtype="complex"
            )

            # loop over unmasked windows
            localWin: int = 0
            global2local: Dict = {}
            for iWin in unmaskedWindows:
                # do the local to global map
                global2local[iWin] = localWin

                # get the window for the input site
                inSF, inReader = self.winSelector.getSpecReaderForWindow(
                    self.inSite, iDec, iWin
                )
                inData = inReader.readBinaryWindowGlobal(iWin).data

                # get the window and channels for the output site
                if self.outSite != self.inSite:
                    outSF, outReader = self.winSelector.getSpecReaderForWindow(
                        self.outSite, iDec, iWin
                    )
                    outData = outReader.readBinaryWindowGlobal(iWin).data
                else:
                    outData = inData

                # get data into the right part of the arrays
                for i in range(0, self.inSize):
                    winDataArray[i] = inData[self.inChannels[i]]
                for i in range(0, self.outSize):
                    winDataArray[self.inSize + i] = outData[self.outChannels[i]]

                # and now can fill the parts of the matrix
                # recall, smooth the power spectra
                for i in range(0, totalSize):
                    for j in range(i, totalSize):
                        winSpectraMatrix[i, j] = smooth1d(
                            winDataArray[i] * np.conjugate(winDataArray[j]),
                            smoothLen,
                            self.win,
                        )
                        if i != j:
                            # due to complex symmetry
                            winSpectraMatrix[j, i] = np.conjugate(
                                winSpectraMatrix[i, j]
                            )

                # after running through all windows, calculate various spectral properties at evaluation frequencies
                # using interpolation
                evalFreqData[:, localWin] = self.calcEvalFrequencyData(
                    freq, evalFreq, winSpectraMatrix
                )

                # increment local window
                localWin = localWin + 1

            # now all the data has been collected
            # for each evaluation frequency, do the robust processing
            # and get the evaluation frequency data
            for eIdx in range(0, numEvalFreq):
                self.printText(
                    "Processing evaluation frequency = {:.6f} [Hz], period = {:.6f} [s]".format(
                        evalFreq[eIdx], 1 / evalFreq[eIdx]
                    )
                )
                # get the constrained windows for the evaluation frequency
                evalFreqWindows = self.winSelector.getWindowsForFreq(iDec, eIdx)
                if len(evalFreqWindows) == 0:  # no windows meet constraints
                    self.printText("No windows found - possibly due to masking")
                    continue
                localWinIndices = []
                for iW in evalFreqWindows:
                    localWinIndices.append(global2local[iW])
                self.printText(
                    "{:d} windows will be solved for".format(len(localWinIndices))
                )
                # restrict processing to data that meets constraints for this evaluation frequency
                # add to class vars
                self.evalFreq.append(evalFreq[eIdx])
                # solution using all components
                numSolveWindows, obs, reg = self.prepareLinearEqn(
                    evalFreqData[eIdx, localWinIndices]
                )
                tmp1, tmp2 = self.robustProcess(numSolveWindows, obs, reg)
                # tmp1, tmp2 = self.olsProcess(numSolveWindows, obs, reg)
                self.impedances.append(tmp1)
                self.variances.append(tmp2)

        if len(self.evalFreq) == 0:
            self.printWarning(
                "No data was found at any decimation level for insite {}, outsite {}, sampling frequency {} and specdir {}".format(
                    self.inSite,
                    self.outSite,
                    self.decParams.getSampleFreqLevel(0),
                    self.winSelector.specdir,
                )
            )
            return

        # write out all the data
        self.writeTF(
            self.winSelector.specdir,
            self.postpend,
            self.evalFreq,
            self.impedances,
            self.variances,
        )

    def getWindowSmooth(self, **kwargs):
        """Window smoothing length

        Power spectra data is smoothed. This returns the size of the window smoother. The window itself is given by win.
    
        Parameters
        ----------
        kwargs['datasize] : int
            The size of the data

        Returns
        -------
        smoothLen : int
            Smoothing size
        """

        # check if window size specified by user
        if self.winSmooth != -1 and self.winSmooth > 1:
            return self.winSmooth
        # if not, calculate based on datasize
        if "datasize" in kwargs:
            winSmooth = kwargs["datasize"] * 1.0 / 16.0
            if winSmooth < 3:
                return 3  # minimum smoothing
            # otherwise round to nearest odd number
            winSmooth = np.ceil(winSmooth) // 2  # this is floor division
            return int(winSmooth * 2 + 1)
        # otherwise, return a default value
        return 15

    def calcEvalFrequencyData(self, freq, evalFreq, winDataMatrix):
        """Calculate spectral power data at evaluation frequencies

        Parameters
        ----------
        freq : np.ndarray
            Frequency array of spectra data
        evalFreq : np.ndarray
            Evaluation frequencies for the decimation level
        winDataMatrix : np.ndarray
            Array holding spectral power data at frequencies freq        

        Returns
        -------
        out : np.ndarray
            Spectral power data interpolated to evaluation frequencies 
        """

        inShape: Tuple = winDataMatrix.shape
        data: np.ndarray = np.empty(
            shape=(evalFreq.size, inShape[0], inShape[1]), dtype="complex"
        )
        # get data from winDataMatrix
        for i in range(0, inShape[0]):
            for j in range(0, inShape[1]):
                interpFunc = interp.interp1d(freq, winDataMatrix[i, j])
                interpVals = interpFunc(evalFreq)
                for eIdx, eFreq in enumerate(evalFreq):
                    data[eIdx, i, j] = interpVals[eIdx]
        return data

    def smoothSpectralEstimates(self, data):
        # takes the evaluation frequency data, which is indexed
        # windows, matrix of spectral components
        winSmooth = 9
        totalChans = self.inSize + self.outSize
        for i in range(0, totalChans):
            for j in range(0, totalChans):
                data[:, i, j] = smooth1d(data[:, i, j], winSmooth, self.win)
        return data

    def checkForBadValues(self, numWindows, data):
        finiteArray = np.ones(shape=(numWindows))
        for iW in range(0, numWindows):
            if not np.isfinite(data[iW]).all():
                finiteArray[iW] = 0
        numGoodWindows = sum(finiteArray)
        if numGoodWindows == numWindows:
            return numWindows, data
        self.printWarning(
            "Bad data found...number of windows reduced from {} to {}".format(
                numWindows, numGoodWindows
            )
        )
        goodWindowIndices = np.where(finiteArray == 1)
        return numGoodWindows, data[goodWindowIndices]

    def prepareLinearEqn(self, data):
        # prepare observations and regressors for linear processing
        numWindows = data.shape[0]
        numWindows, data = self.checkForBadValues(numWindows, data)
        crossSize = len(self.crossChannels)
        # for each output variable, have ninput regressor variables
        # construct our arrays
        obs = np.empty(shape=(self.outSize, crossSize * numWindows), dtype="complex")
        reg = np.empty(
            shape=(self.outSize, crossSize * numWindows, self.inSize), dtype="complex"
        )
        for iW in range(0, numWindows):
            iOffset = iW * crossSize
            for i in range(0, self.outSize):
                for j, crossChan in enumerate(self.crossChannels):
                    # this is the observation row where i is the observed output
                    crossIndex = self.allChannels.index(crossChan)
                    obs[i, iOffset + j] = data[iW, self.inSize + i, crossIndex]
                    for k in range(0, self.inSize):
                        reg[i, iOffset + j, k] = data[iW, k, crossIndex]
        return numWindows, obs, reg

    # SOLVER ROUTINES
    def robustProcess(self, numWindows, obs, reg):
        # do the robust processing for a single evaluation frequency
        crossSize = len(self.crossChannels)
        # create array for output
        output = np.empty(shape=(self.outSize, self.inSize), dtype="complex")
        varOutput = np.empty(shape=(self.outSize, self.inSize), dtype="float")
        # solve
        for i in range(0, self.outSize):
            observation = obs[i, :]
            predictors = reg[i, :, :]
            # save the output
            out, resids, weights = chatterjeeMachler(
                predictors, observation, intercept=self.intercept
            )
            # out, resids, scale, weights = mmestimateModel(predictors, observation, intercept=self.intercept)

            # now take the weights, apply to the observations and predictors, stack the appropriate rows
            observation2 = np.zeros(shape=(crossSize), dtype="complex")
            predictors2 = np.zeros(shape=(crossSize, self.inSize), dtype="complex")
            for iChan in range(0, crossSize):
                # now need to have my indexing array
                indexArray = np.arange(iChan, numWindows * crossSize, crossSize)
                weightsLim = weights[indexArray]
                # weightsLim = weightsLim/np.sum(weightsLim) # normalise weights to 1
                observation2[iChan] = (
                    np.sum(obs[i, indexArray] * weightsLim) / numWindows
                )
                # now for the regressors
                for j in range(0, self.inSize):
                    predictors2[iChan, j] = (
                        np.sum(reg[i, indexArray, j] * weightsLim) / numWindows
                    )
            out, resids, weights = chatterjeeMachler(
                predictors2, observation2, intercept=self.intercept
            )
            # out, resids, scale, weights = mmestimateModel(
            #     predictors2, observation2, intercept=self.intercept)

            # now calculate out the varainces - have the solution out, have the weights
            # recalculate out the residuals with the final solution
            # calculate standard deviation of residuals
            # and then use chatterjee machler formula to estimate variances
            # this needs work - better to use an empirical bootstrap method, but this will do for now
            resids = np.absolute(observation - np.dot(predictors, out))
            scale = sampleMAD0(
                resids
            )  # some measure of standard deviation, rather than using the standard deviation
            residsVar = scale * scale
            # varPred = np.dot(hermitianTranspose(predictors), weights*predictors) # need to fix this
            varPred = np.dot(hermitianTranspose(predictors), predictors)
            varPred = np.linalg.inv(varPred)  # this is a pxp matrix
            varOut = 1.91472 * residsVar * varPred
            varOut = np.diag(varOut).real  # this should be a real number
            if self.intercept:
                output[i] = out[1:]
                varOutput[i] = varOut[1:]
            else:
                output[i] = out
                varOutput[i] = varOut

        return output, varOutput

    def olsProcess(self, numWindows, obs, reg):
        # ordinary least squares solution
        # create array for output
        output = np.empty(shape=(self.outSize, self.inSize), dtype="complex")
        varOutput = np.empty(shape=(self.outSize, self.inSize), dtype="float")
        # solve
        for i in range(0, self.outSize):
            observation = obs[i, :]
            predictors = reg[i, :, :]
            # save the output
            out, resids, squareResid, rank, s = olsModel(
                predictors, observation, intercept=self.intercept
            )
            # if self.intercept:
            # 	output[i] = out[1:]
            # else:
            # 	output[i] = out

            # now calculate out the varainces - have the solution out, have the weights
            # recalculate out the residuals with the final solution
            # calculate standard deviation of residuals
            # and then use chatterjee machler formula to estimate variances
            # this needs work - better to use an empirical bootstrap method, but this will do for now
            resids = np.absolute(observation - np.dot(predictors, out))
            scale = sampleMAD0(
                resids
            )  # some measure of standard deviation, rather than using the standard deviation
            residsVar = scale * scale
            # varPred = np.dot(hermitianTranspose(predictors), weights*predictors) # need to fix this
            varPred = np.dot(hermitianTranspose(predictors), predictors)
            varPred = np.linalg.inv(varPred)  # this is a pxp matrix
            varOut = 1.91472 * residsVar * varPred
            varOut = np.diag(varOut).real  # this should be a real number
            if self.intercept:
                output[i] = out[1:]
                varOutput[i] = varOut[1:]
            else:
                output[i] = out
                varOutput[i] = varOut

        return output, varOutput

    def stackedProcess(self, data):
        # then do various sums
        numWindows = data.shape[0]
        crossSize = len(self.crossChannels)
        # unweighted sum (i.e. normal solution)
        unWeightedSum = np.sum(data, axis=0)
        unWeightedSum = unWeightedSum / numWindows

        # for each output variable, have ninput regressor variables
        # let's construct our arrays
        obs = np.empty(shape=(self.outSize, crossSize), dtype="complex")
        reg = np.empty(shape=(self.outSize, crossSize, self.inSize), dtype="complex")
        for i in range(0, self.outSize):
            for j, crossChan in enumerate(self.crossChannels):
                crossIndex = self.allChannels.index(crossChan)
                obs[i, j] = unWeightedSum[self.inSize + i, crossIndex]
                for k in range(0, self.inSize):
                    reg[i, j, k] = unWeightedSum[k, crossIndex]

        # create array for output
        output = np.empty(shape=(self.outSize, self.inSize), dtype="complex")

        for i in range(0, self.outSize):
            observation = obs[i, :]
            predictors = reg[i, :, :]
            # save the output
            out, resids, scale, weights = mmestimateModel(
                predictors, observation, intercept=self.intercept
            )
            if self.intercept:
                output[i] = out[1:]
            else:
                output[i] = out
        return output

    def writeTF(self, specdir: str, postpend: str, freq, data, variances, **kwargs):
        # path for writing out to
        sampleFreqStr = fileFormatSampleFreq(self.decParams.sampleFreq)
        if postpend == "":
            filename = "{}_fs{:s}_{}".format(self.outSite, sampleFreqStr, specdir)
        else:
            filename = "{}_fs{:s}_{}_{}".format(
                self.outSite, sampleFreqStr, specdir, postpend
            )
        datapath = os.path.join(self.outpath, sampleFreqStr)
        checkAndMakeDir(datapath)
        outfile = os.path.join(datapath, filename)
        # now construct the transferFunctionData object
        numFreq = len(freq)
        dataDict = {}
        varDict = {}
        for i in range(0, self.outSize):
            for j in range(0, self.inSize):
                key = "{}{}".format(self.outChannels[i], self.inChannels[j])
                dataArray = np.empty(shape=(numFreq), dtype="complex")
                varArray = np.empty(shape=(len(freq)), dtype="float")
                for ifreq in range(0, numFreq):
                    dataArray[ifreq] = data[ifreq][i, j]
                    varArray[ifreq] = variances[ifreq][i, j]
                dataDict[key] = dataArray
                varDict[key] = varArray
        tfData = TransferFunctionData(freq, dataDict, varDict)
        # now make the writer and write out
        tfWriter = TransferFunctionWriter(outfile, tfData)
        tfWriter.setHeaders(
            sampleFreq=self.decParams.sampleFreq,
            insite=self.inSite,
            inchans=self.inChannels,
            outsite=self.outSite,
            outchans=self.outChannels,
        )
        if "remotesite" in kwargs:
            tfWriter.addHeader("remotesite", kwargs["remotesite"])
        if "remotechans" in kwargs:
            tfWriter.addHeader("remotechans", kwargs["remotechans"])
        tfWriter.write()

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append("In Site = {}".format(self.inSite))
        textLst.append("In Channels = {}".format(self.inChannels))
        textLst.append("Out Site = {}".format(self.outSite))
        textLst.append("Out Channels = {}".format(self.outChannels))
        return textLst



import numpy as np
import math
import xml.etree.ElementTree as ET
from typing import List, Tuple


def getKnownCalibrationFormats() -> Tuple[List, List]:
    """Return list of supported calibration formats and their extensions

    Returns
    -------
    extensions : List
        List of calibration file extensions
    formats : List
        List of calibration file formats 
    """

    calExt = ["TXT", "TXT", "RSP", "RSPX"]
    calFormats = ["induction", "metronix", "rsp", "rspx"]
    return calExt, calFormats


def getCalName(format, ext: str, sensor: str, serial: int, chopper) -> str:
    """Get the calibration file name
    
    Parameters
    ----------
    format : str
        Calibration format
    ext : str
        Calibration file extension
    sensor : str
        Sensor name
    serial : int
        The sensor serial number
    chopper : bool
        Boolean flag for chopper on or off

    Returns
    -------
    out : str
        Name of calibration file
    """

    if format == "induction":
        return inductionName(ext, sensor, serial, chopper)
    elif format == "metronix":
        return metronixName(ext, sensor, serial, chopper)
    elif format == "rsp":
        return rspName(ext, sensor, serial, chopper)
    elif format == "rspx":
        return rspxName(ext, sensor, serial, chopper)
    else:
        return metronixName(ext, sensor, serial, chopper)


def inductionName(ext: str, sensor: str, serial: int, chopper: bool) -> str:
    """Get internal format induction coil calibration file name
    
    Parameters
    ----------
    ext : str
        Calibration file extension
    sensor : str
        Sensor name
    serial : int
        The sensor serial number
    chopper : bool
        Boolean flag for chopper on or off

    Returns
    -------
    out : str
        Name of calibration file
    """

    return "IC_{}.{}".format(serial, ext)


def metronixName(ext: str, sensor: str, serial: int, chopper: bool) -> str:
    """Get Metronix calibration file name
    
    Parameters
    ----------
    ext : str
        Calibration file extension
    sensor : str
        Sensor name
    serial : int
        The sensor serial number
    chopper : bool
        Boolean flag for chopper on or off

    Returns
    -------
    out : str
        Name of calibration file
    """

    return "{}{}.{}".format(sensor, serial, ext)


def rspName(ext: str, sensor: str, serial: int, chopper: bool) -> List[str]:
    """Get RSP calibration file name
    
    Parameters
    ----------
    ext : str
        Calibration file extension
    sensor : str
        Sensor name
    serial : int
        The sensor serial number
    chopper : bool
        Boolean flag for chopper on or off

    Returns
    -------
    out : List[str]
        Name of calibration files
    """

    if len(sensor) < 6:
        # not possible to get a sensor number
        return None

    board = "HF"
    if chopper:
        board = "LF"
    sensorNum = int(sensor[3:5])
    names = []
    names.append("TYPE-{:03d}_{}-ID-{:06d}.{}".format(sensorNum, board, serial, ext))
    names.append("TYPE-{:03d}_BB-ID-{:06d}.{}".format(sensorNum, serial, ext))
    return names


def rspxName(ext: str, sensor: str, serial: int, chopper: bool) -> List[str]:
    """Get RSPX calibration file name
    
    Parameters
    ----------
    ext : str
        Calibration file extension
    sensor : str
        Sensor name
    serial : int
        The sensor serial number
    chopper : bool
        Boolean flag for chopper on or off

    Returns
    -------
    out : List[str]
        Name of calibration files
    """

    if len(sensor) < 6:
        # not possible to get a sensor number
        return None

    board = "HF"
    if chopper:
        board = "LF"
    sensorNum = int(sensor[3:5])
    names = []
    names.append("TYPE-{:03d}_{}-ID-{:06d}.{}".format(sensorNum, board, serial, ext))
    names.append("TYPE-{:03d}_BB-ID-{:06d}.{}".format(sensorNum, serial, ext))
    return names


def defaultCalibration():
    """Default calibration data

    Returns
    -------
    data : np.ndarray
        Data lines converted to a float array
    staticGain : float
        Static gain    
    """

    return [1] * 10, 1



import os
from datetime import datetime, timedelta
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
from typing import List, Union

# import from package
from resistics.dataObjects.dataObject import DataObject
from resistics.utilities.utilsChecks import isElectric
from resistics.utilities.utilsPlotter import getViewFonts


class TimeData(DataObject):
    """Class for holding time data

    Attributes
    ----------
    numSamples : int
        The number of samples in the data
    sampleFreq : float
        The sampling frequency
    startTime : datetime.datetime
        The time of the first sample
    stopTime : datetime.datetime
        The time of the last sample 
    chans : List[str]
        The channels in the data
    numChans : int
        The number of channels
    data : Dict
        The time data with channels as keys and arrays as values
    comments : List[str]
        Information about the time data as a list of strings

    Methods
    -------
    __init__(kwargs)
        Initialise the time data
    setData(windowSize, dataSize, sampleFreq, startTime, stopTime, data)
        Set data with parameters
    getDateArray() : np.ndarray
        A datetime array of the sample times
    getComments()
        Get a deepcopy of the comments
    addComment(comment)
        Add a comment to the dataset
    copy()
        Get a copy of the timeseries data
    view(kwargs)
        View the time data 
    printList()
        Class status returned as list of strings          
    """

    def __init__(
        self,
        sampleFreq: float,
        startTime: Union[datetime, str],
        stopTime: Union[datetime, str],
        data,
        comments: Union[str, List[str]] = [],
    ) -> None:
        """Initialise and set object parameters

        Parameters
        ----------
        sampleFreq : float
            The sampling frequency in Hz
        startTime : datetime, str
            The startTime of the window
        stopTime : datetime, str
            The stopTime of the window 
        data : Dict
            The data dictionary with keys as channels and values as spectra data
        comments : List[str]
            Dataset comments
        """

        self.setData(sampleFreq, startTime, stopTime, data)
        self.comments = comments
        if isinstance(self.comments, str):
            self.comments = [self.comments]

    def setData(
        self,
        sampleFreq: float,
        startTime: Union[datetime, str],
        stopTime: Union[datetime, str],
        data,
    ):
        """Set the object parameters

        Parameters
        ----------
        sampleFreq : float
            The sampling frequency in Hz
        startTime : datetime, str
            The startTime of the window
        stopTime : datetime, str
            The stopTime of the window 
        data : Dict
            The data dictionary with keys as channels and values as spectra data
        """

        self.sampleFreq = sampleFreq
        # start time
        self.startTime = (
            datetime.strptime(startTime, "%Y-%m-%d %H:%M:%S.%f")
            if isinstance(startTime, str)
            else startTime
        )
        # stop time
        self.stopTime = (
            datetime.strptime(stopTime, "%Y-%m-%d %H:%M:%S.%f")
            if isinstance(stopTime, str)
            else stopTime
        )
        # other properties
        self.chans = sorted(data.keys())
        self.numChans = len(self.chans)
        self.data = data
        self.numSamples = data[self.chans[0]].size

    def getDateArray(self) -> np.ndarray:
        """Get the date array

        Returns
        -------
        np.ndarray
            The date array of the time samples
        """

        x = np.empty(shape=(self.numSamples), dtype=datetime)
        for i in range(0, self.numSamples):
            x[i] = self.startTime + timedelta(seconds=1.0 * i / self.sampleFreq)
        return x

    def getComments(self) -> List[str]:
        """Get a deepcopy of the comments
        
        Returns
        -------
        List[str]
            Dataset comments as a list of strings
        """

        return deepcopy(self.comments)

    def addComment(self, comment: str) -> None:
        """Add a new comment

        Parameters
        ----------
        comment : str
            A new comment
        """

        self.comments.append(comment)

    def copy(self):
        """Get a copy of the time data object

        Returns
        -------
        TimeData
            A copy of the time data object
        """

        return TimeData(
            self.sampleFreq,
            self.startTime,
            self.stopTime,
            deepcopy(self.data),
            self.getComments(),
        )

    def view(self, **kwargs) -> plt.figure:
        """Plot statistics for evaluation frequency index

        Plots a simple scatter of each statistic with datetime on the xaxis (datetime of the window start dates). Number of subplots is equal to numStaStatPerWindow.

        Parameters
        ----------
        sampleStart : int, optional
            Sample to start plotting from
        sampleStop : int, optional
            Sample to plot to                   
        fig : matplotlib.pyplot.figure, optional
            A figure object
        plotfonts : Dict, optional
            A dictionary of plot fonts
        chans : List[str]
            Channels to plot
        label : str, optional
            Label for the plots
        xlim : List, optional
            Limits for the x axis
        legened : bool
            Boolean flag for adding a legend
        
        Returns
        -------
        plt.figure
            Matplotlib figure object
        """

        # the number of samples to plot
        sampleStart = 0
        sampleStop = 4096
        if "sampleStart" in kwargs:
            sampleStart = kwargs["sampleStart"]
        if "sampleStop" in kwargs:
            sampleStop = kwargs["sampleStop"]
        if sampleStop >= self.numSamples:
            sampleStop = self.numSamples - 1
        # get the x axis ready
        x = self.getDateArray()
        start = x[sampleStart]
        stop = x[sampleStop]

        # now plot
        fig = (
            plt.figure(kwargs["fig"].number)
            if "fig" in kwargs
            else plt.figure(figsize=(20, 2 * self.numChans))
        )
        plotfonts = kwargs["plotfonts"] if "plotfonts" in kwargs else getViewFonts()
        # suptitle
        st = fig.suptitle(
            "Time data from {} to {}, samples {} to {}".format(
                start, stop, sampleStart, sampleStop
            ),
            fontsize=plotfonts["suptitle"],
        )
        st.set_y(0.98)
        # now plot the data
        dataChans = kwargs["chans"] if "chans" in kwargs else self.chans
        numDataChans = len(dataChans)
        for idx, chan in enumerate(dataChans):
            ax = plt.subplot(numDataChans, 1, idx + 1)
            # title and label for the plot
            plt.title("Channel {}".format(chan), fontsize=plotfonts["title"])
            lab = (
                kwargs["label"]
                if "label" in kwargs
                else "{}: {} to {}".format(chan, start, stop)
            )
            # plot data
            plt.plot(
                x[sampleStart : sampleStop + 1],
                self.data[chan][sampleStart : sampleStop + 1],
                label=lab,
            )
            # add time label
            if idx == numDataChans - 1:
                plt.xlabel("Time", fontsize=plotfonts["axisLabel"])
            # set the xlim
            xlim = kwargs["xlim"] if "xlim" in kwargs else [start, stop]
            plt.xlim(xlim)
            # y axis options
            if isElectric(chan):
                plt.ylabel("mV/km", fontsize=plotfonts["axisLabel"])
            else:
                plt.ylabel("nT or mV", fontsize=plotfonts["axisLabel"])
            plt.grid()
            # set tick sizes
            for label in ax.get_xticklabels() + ax.get_yticklabels():
                label.set_fontsize(plotfonts["axisTicks"])
            # legend
            if "legend" in kwargs and kwargs["legend"]:
                plt.legend(loc=4)

        # show if the figure is not in keywords
        if "fig" not in kwargs:
            plt.tight_layout(rect=[0, 0.02, 1, 0.96])
            plt.show()
        
        return fig

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Sampling frequency [Hz] = {}".format(self.sampleFreq))
        textLst.append("Sample rate [s] = {}".format(1.0 / self.sampleFreq))
        textLst.append("Number of samples = {}".format(self.numSamples))
        textLst.append("Number of channels = {}".format(self.numChans))
        textLst.append("Channels = {}".format(self.chans))
        textLst.append("Start time = {}".format(self.startTime))
        textLst.append("Stop time = {}".format(self.stopTime))
        if len(self.comments) == 0:
            textLst.append("No comments")
        else:
            textLst.append("Comments")
            for comment in self.comments:
                textLst.append("\t{}".format(comment))
        return textLst
